import scipy as SP
import h5py
import copy
import data_util as du
try:
    #see if fastlmm is in the path for the fast C-based parser
    import fastlmm.pyplink.wrap_plink_parser as parser  
    cparser=True
except:
    cparser=False
    pass

def estCumPos(pos,chrom,offset = 20000000):
    '''
    compute the cumulative position of each variant given the position and the chromosome
    Also return the starting cumulativeposition of each chromosome

    Args:
        pos:        scipy.array of basepair positions (on the chromosome)
        chrom:      scipy.array of chromosomes
        offset:     offset between chromosomes for cumulative position (default 20000000 bp)
    
        Returns:
        cum_pos:    scipy.array of cumulative positions
        chrom_pos:  scipy.array of starting cumulative positions for each chromosme
    '''
    chromvals = SP.unique(chrom)#SP.unique is always sorted
    chrom_pos=SP.zeros_like(chromvals)#get the starting position of each Chrom
    cum_pos = SP.zeros_like(pos)#get the cum_pos of each variant.
    maxpos_cum=0
    for i,mychrom in enumerate(chromvals):
        chrom_pos[i] = maxpos_cum
        i_chr=chrom==mychrom
        maxpos = pos[i_chr].max()+offset
        maxpos_cum+=maxpos
        cum_pos[i_chr]=chrom_pos[i]+pos[i_chr]
    return cum_pos,chrom_pos
        
        
def _imputeMissing(X, center=True, unit=True, betaNotUnitVariance=False, betaA=1.0, betaB=1.0):
        '''
        fill in missing values in the SNP matrix by the mean value
        optionally center the data and unit-variance it

        Args:
            X:      scipy.array of SNP values. If dtype=='int8' the missing values are -9, 
                    otherwise the missing values are scipy.nan
            center: Boolean indicator if data should be mean centered
                    Not supported in C-based parser
            unit:   Boolean indicator if data should be normalized to have unit variance
                    Not supported in C-based parser
            betaNotUnitVariance:    use Beta(betaA,betaB) standardization instead of unit variance 
                                    (only with C-based parser) (default: False)
            betaA:  shape parameter for Beta(betaA,betaB) standardization (only with C-based parser)
            betaB:  scale parameter for Beta(betaA,betaB) standardization (only with C-based parser)

        Returns:
            X:      scipy.array of standardized SNPs with scipy.float64 values
        '''
        typeX=X.dtype
        if typeX!=SP.int8:
            iNanX = X!=X
        else:
            iNanX = X==-9
        if iNanX.any() or betaNotUnitVariance:
            if cparser:
                print("using C-based imputer")
                if X.flags["C_CONTIGUOUS"] or typeX!=SP.float32:
                    X = SP.array(X, order="F", dtype=SP.float32)
                    if typeX==SP.int8:
                        X[iNanX]=SP.nan
                    parser.standardize(X,betaNotUnitVariance=betaNotUnitVariance,betaA=betaA,betaB=betaB)
                    X=SP.array(X,dtype=SP.float64)
                else:
                    parser.standardize(X,betaNotUnitVariance=betaNotUnitVariance,betaA=betaA,betaB=betaB)
                X=SP.array(X,dtype=SP.float64)
            else:
                if betaNotUnitVariance:
                    raise NotImplementedError("Beta(betaA,betaB) standardization only in C-based parser, but not found")
                nObsX = (~iNanX).sum(0)
                if typeX!=SP.float64:
                    X=SP.array(X,dtype=SP.float64)
                X[iNanX] = 0.0
                sumX = (X).sum(0)                
                meanX = sumX/nObsX
                if center:
                    X-=meanX
                    X[iNanX] = 0.0
                    X_=X
                else:
                    mean=SP.tile(meanX,(X.shape[0],1))
                    X[iNanX]=mean[iNanX]
                    X_=X-mean
                if unit:
                    stdX = SP.sqrt((X_*X_).sum(0)/nObsX)
                    stdX[stdX==0.0]=1.0
                    X/=stdX
        else:
            if X.dtype!=SP.float64:
                X=SP.array(X,dtype=SP.float64)
            if center:
                X-= X.mean(axis=0)
            if unit:
                stdX= X.std(axis=0)
                stdX[stdX==0.0]=1.0
                X/=stdX
        return X

class QTLData():
    def __init__(self,file_name):
        self.file_name = file_name
        self.load()


    def load(self,cache_genotype=False,cache_phenotype=True):
        """load data file
        
        Args:
            cache_genotype:     load genotypes fully into memory (default: False)
            cache_phenotype:    load phentopyes fully intro memry (default: True)        
        """
        self.f = h5py.File(self.file_name,'r')
        self.pheno = self.f['phenotype']
        self.geno  = self.f['genotype']
        #TODO: load all row and column headers for genotype and phenotype

        #parse out thse we alwasy need for convenience
        self.genoM = self.geno['matrix']

        self.phenoM = self.pheno['matrix']
        self.sample_ID = self.geno['row_header']['sample_ID'][:]
        self.genoChrom = self.geno['col_header']['chrom'][:]
        self.genoPos   = self.geno['col_header']['pos'][:]
        if 'pos_cum' in list(self.geno['col_header'].keys()):
            self.genoPos_cum   = self.geno['col_header']['pos_cum'][:]
        else:
            self.genoPos_cum = None
        self.phenotype_ID = self.pheno['col_header']['phenotype_ID'][:]


        #cache?
        if cache_genotype:
            self.genoM = self.genoM[:]
        if cache_phenotype:
            self.phenoM = self.phenoM[:]
    
        # Additional pheno col header
        headers = list(self.pheno['col_header'].keys())
        if 'gene_ID' in headers:
            self.eqtl = True
            self.geneID = self.pheno['col_header']['gene_ID'][:]
            self.gene_pos = SP.array([self.pheno['col_header']['gene_chrom'][:],self.pheno['col_header']['gene_start'][:],self.pheno['col_header']['gene_end']],dtype='int').T
            self.geneIDs= list(set(self.geneID))
        else:
            self.eqtl = False
        if 'environment' in headers:
            self.E  = self.pheno['col_header/environment'][:]
            self.Es = list(set(self.E))
        else:
            self.E = None

        #dimensions
        self.N = self.genoM.shape[0]
        self.S = self.genoM.shape[1]
        self.P = self.phenoM.shape[1]
        assert (self.genoM.shape[0]==self.phenoM.shape[0]), 'dimension missmatch'

    def getGenoIndex(self,pos0=None,pos1=None,chrom=None,pos_cum0=None,pos_cum1=None):
        """computes 0-based genotype index from position of cumulative position. 
        Positions can be given in one out of two ways: 
        - position (pos0-pos1 on chrom)
        - cumulative position (pos_cum0-pos_cum1)
        If all these are None (default), then all genotypes are returned

        Args:
            pos0:       position based selection (start position)
            pos1:       position based selection (stop position)
            chrom:      position based selection (chromosome)
            pos_cum0:   cumulative position based selection (start position)
            pos_cum1:   cumulative position based selection (stop position)
        
        Returns:
            i0:         genotype index based selection (start index)
            i1:         genotype index based selection (stop index)
        """
        if (pos0 is not None) & (pos1 is not None) & (chrom is not None):
            I = self.gneoChrom==chrom
            I = I & (self.genoPos>=p0) & (self.genoPos<p1)
            I = SP.nonzero(I)[0]
            i0 = I.min()
            i1 = I.max()
        elif (pos_cum0 is not None) & (pos_cum1 is not None):
            I = (self.genoPos_cum>=pos_cum0) & (self.genoPos_cum<pos_cum1)
            I = SP.nonzero(I)[0]
            if I.size==0:
                return None
            i0 = I.min()
            i1 = I.max()
        else:
            i0=None
            i1=None
        return i0,i1
 
    def getGenotypes(self,i0=None,i1=None,pos0=None,pos1=None,chrom=None,center=True,unit=True,pos_cum0=None,pos_cum1=None,impute_missing=True):
        """load genotypes. 
        Optionally the indices for loading subgroups the genotypes for all people
        can be given in one out of three ways: 
        - 0-based indexing (i0-i1)
        - position (pos0-pos1 on chrom)
        - cumulative position (pos_cum0-pos_cum1)
        If all these are None (default), then all genotypes are returned

        Args:
            i0:         genotype index based selection (start index)
            i1:         genotype index based selection (stop index)
            pos0:       position based selection (start position)
            pos1:       position based selection (stop position)
            chrom:      position based selection (chromosome)
            pos_cum0:   cumulative position based selection (start position)
            pos_cum1:   cumulative position based selection (stop position)
            impute_missing: Boolean indicator variable if missing values should be imputed
        
        Returns:
            X:          scipy.array of genotype values
        """
        #position based matching?
        if (i0 is None) and (i1 is None) and ((pos0 is not None) & (pos1 is not None) & (chrom is not None)) or ((pos_cum0 is not None) & (pos_cum1 is not None)):
            i0,i1=self.getGenoIndex(pos0=pos0,pos1=pos1,chrom=chrom,pos_cum0=pos_cum0,pos_cum1=pose_cum1)
        #index based matching?
        if (i0 is not None) & (i1 is not None):
            X = self.genoM[:,i0:i1]
        else:
            X = self.genoM[:,:]
        if impute_missing:
            X = du.imputeMissing(X,center=center,unit=unit)
        return X

    def getCovariance(self,normalize=True,i0=None,i1=None,pos0=None,pos1=None,chrom=None,center=True,unit=True,pos_cum0=None,pos_cum1=None,blocksize=None,X=None,**kw_args):
        """calculate the empirical genotype covariance in a region"""
        if X is not None:
            K=X.dot(X.T)
            Nsnp=X.shape[1]
        else:
            if (i0 is None) and (i1 is None) and ((pos0 is not None) & (pos1 is not None) & (chrom is not None)) or ((pos_cum0 is not None) & (pos_cum1 is not None)):
                i0,i1=self.getGenoIndex(pos0=pos0,pos1=pos1,chrom=chrom,pos_cum0=pos_cum0,pos_cum1=pose_cum1)

            [N,M]=self.genoM.shape
            if blocksize is None:
                blocksize=M
            if i0 is None:
                i0=0
            if i1 is None:
                i1=M
            nread = i0
            K=None
            Nsnp=i1-i0
            while nread<i1:
                thisblock=min(blocksize,i1-nread)
                X=self.getGenotypes(i0=nread,i1=(nread+thisblock),center=center,unit=unit,**kw_args)    
                if K is None:
                    K=X.dot(X.T)
                else:
                    K+=X.dot(X.T)
                nread+=thisblock
        if normalize:
            K/=(K.diagonal().mean())
        else:#divide by number of SNPs in K
            K/=Nsnp
        return K

    def getGenoID(self,i0=None,i1=None,pos0=None,pos1=None,chrom=None,pos_cum0=None,pos_cum1=None):
        """get genotype IDs. 
        Optionally the indices for loading subgroups the genotype IDs for all people
        can be given in one out of three ways: 
        - 0-based indexing (i0-i1)
        - position (pos0-pos1 on chrom)
        - cumulative position (pos_cum0-pos_cum1)
        If all these are None (default), then all genotypes are returned

        Args:
            i0:         genotype index based selection (start index)
            i1:         genotype index based selection (stop index)
            pos0:       position based selection (start position)
            pos1:       position based selection (stop position)
            chrom:      position based selection (chromosome)
            pos_cum0:   cumulative position based selection (start position)
            pos_cum1:   cumulative position based selection (stop position)
           
        Returns:
            ID:         scipy.array of genotype IDs (e.g. rs IDs)
        """
        #position based matching?
        if (i0 is None) and (i1 is None) and ((pos0 is not None) & (pos1 is not None) & (chrom is not None)) or ((pos_cum0 is not None) & (pos_cum1 is not None)):
            i0,i1=self.getGenoIndex(pos0=pos0,pos1=pos1,chrom=chrom,pos_cum0=pos_cum0,pos_cum1=pose_cum1)
        if "genotype_id" in list(self.geno.keys()):
            if (i0 is not None) & (i1 is not None):
                return self.geno["genotype_id"][i0:i1]
            else:
                return self.geno["genotype_id"][i0:i1]
        else:
            if (i0 is not None) & (i1 is not None):
                return SP.arange(i0,i0)
            else:
                return SP.arange(self.genoM.shape[1])
        pass

    def getPhenotypes(self,i0=None,i1=None,phenotype_IDs=None,geneIDs=None,environments=None,center=True,impute=True,intersection=False):
        """load Phenotypes
        
        Args:
            i0:             phenotype indices to load (start individual index)
            i1:             phenotype indices to load (stop individual index)
            phenotype_IDs:  names of phenotypes to load
            geneIDs:        names of genes to load
            environments:   names of environments to load
            impute:         imputation of missing values (default: True)
            intersection:   restrict observation to those obseved in all phenotypes? (default: False)
        
        Returns:
            Y:              phenotype values
            Ikeep:          index of individuals in Y
        """
        if phenotype_IDs is None and (geneIDs is not None or environments is not None):
           if geneIDs is None:
               geneIDs = self.geneIDs
           elif type(geneIDs)!=list:
               geneIDs = [geneIDs]
           if environments is None:
               environments = self.Es
           elif type(environments)!=list:
               environments = [environments]
           phenotype_IDs = []
           for env in environments:
               for gene in geneIDs:
                   phenotype_IDs.append('%s:%d'%(gene,env))

        if phenotype_IDs is not None:
            I = SP.array([SP.nonzero(self.phenotype_ID==n)[0][0] for n in phenotype_IDs])
        else:
            I = SP.arange(self.phenotype_ID.shape[0])
        Y = SP.array(self.phenoM[:,I],dtype='float')
        Iok = (~SP.isnan(Y))

        if intersection:
            Ikeep = Iok.all(axis=1)
        else:
            Ikeep = Iok.any(axis=1)

        Y = Y[Ikeep]
        Iok = Iok[Ikeep]

        if center | impute:
            for i in range(Y.shape[1]):
                ym = Y[Iok[:,i],i].mean()
                Y[:,i]-=ym
                Y[~Iok[:,i],i] = ym
                Y[:,i]/=Y[:,i].std()
        #calculate overlap of missing values
        return [Y,Ikeep]

    def getPos(self):
        """
        get the positions of the genotypes

        Returns:
            chromosome
            position
            cumulative_position
        """
        return [self.genoChrom,self.genoPos,self.genoPos_cum]

    def getIcis_geno(self,geneID,cis_window=50E3):
        """ if eqtl==True it returns a bool vec for cis """
        assert self.eqtl == True, 'Only for eqtl data'
        index = self.geneID==geneID
        [_chrom,_gene_start,_gene_end] = self.gene_pos[index][0,:]
        Icis = (self.genoChrom==_chrom)*(self.genoPos>=_gene_start-cis_window)*(self.genoPos<=_gene_end+cis_window)
        return Icis

    def subSample(self,Irow=None,Icol_geno=None,Icol_pheno=None):
        """sample a particular set of individuals (Irow) or phenotypes (Icol_pheno) or genotypes (Icol_geno)
        
        Args:
            Irow:           indices for a set of individuals
            Icol_pheno:     indices for a set of phenotypes
            Icol_geno:      indices for a set of SNPs
        
        Returns:
            QTLdata opject holding the specified subset of the data
        """
        C = copy.copy(self)
        if Irow is not None:
            C.genoM = C.genoM[Irow]
            C.phenoM = C.phenoM[Irow]
            C.sample_ID = C.sample_ID[Irow]
        if Icol_geno is not None:
            C.genoM = C.genoM[:,Icol_geno]
            C.genoPos = C.genoPos[Icol_geno]
            C.genoChrom = C.genoChrom[Icol_geno]
            C.genoPos_cum = C.genoPos_cum[Icol_geno]
        if Icol_pheno is not None:
            C.phenoM = C.phenoM[:,Icol_pheno]
            C.phenotype_ID = C.phenotype_ID[Icol_pheno]
            if C.eqtl:
                C.geneID   = C.geneID[Icol_pheno]
                C.gene_pos = C.gene_pos[Icol_pheno,:]
                C.geneIDs  = list(set(C.geneID))
            if C.E != None:
                C.E  = C.E[Icol_pheno]
                C.Es = list(set(C.E))

        C.N = C.genoM.shape[0]
        C.S = C.genoM.shape[1]
        C.P = C.phenoM.shape[1]
        return C




#! /usr/bin/env python
# Copyright(c) 2014, The mtSet developers (Francesco Paolo Casale, Barbara Rakitsch, Oliver Stegle)
# All rights reserved.

from limix.mtSet.core.simPhenoCore import simPheno
from optparse import OptionParser
import scipy as SP

def entry_point():

    parser = OptionParser()
    parser.add_option("--bfile",     dest='bfile',      type=str, default=None)
    parser.add_option("--cfile",     dest='cfile',      type=str, default=None)
    parser.add_option("--pfile",     dest='pfile',      type=str, default=None)

    parser.add_option("--seed",      dest='seed',       type=int, default=0)
    parser.add_option("--nTraits",   dest='nTraits',    type=int, default=4)
    parser.add_option("--windowSize",dest='windowSize', type=int, default=3e4)

    parser.add_option("--chrom", dest='chrom',  type=int, default=None)
    parser.add_option("--minPos",dest='pos_min',type=int, default=None)
    parser.add_option("--maxPos",dest='pos_max',type=int, default=None)

    parser.add_option("--vTotR",   dest='vTotR', type=float, default=0.05)
    parser.add_option("--nCausalR",dest='nCausalR', type=int,default=8)
    parser.add_option("--pCommonR",dest='pCommonR', type=float,default=0.5)
    parser.add_option("--vTotBg",  dest='vTotBg',type=float,default=0.4)
    parser.add_option("--pHidden", dest='pHidden',type=float,default=0.6)
    parser.add_option("--pCommon", dest='pCommon', type=float,default=0.5)
    (options, args) = parser.parse_args()

    simPheno(options)



import sys
sys.path.append('./..')
sys.path.append('./../../..')

import scipy as SP
import scipy.linalg
import limix
import pdb
import time

#TODO: what is this script? Can we remove for the release?
#development stuff should go into separate branches


def kroneckerApprox(S,K,Gamma):
    
    import scipy.optimize
    
    def fitKronApprox(a):
        Sbg = SP.zeros_like(S[0])
        Kbg = SP.zeros_like(K[0])
        for i in range(len(S)):  Sbg+= a[i]*S[i]
        for i in range(len(K)):  Kbg+= a[i+len(S)]*K[i]
        Gamma1 = SP.kron(Sbg,Kbg)
        return ((Gamma-Gamma1)**2).sum()

    a0 = SP.concatenate((1/float(len(S))*SP.ones(len(S)),1/float(len(K))*SP.ones(len(K))))
    res = SP.optimize.fmin_l_bfgs_b(fitKronApprox,a0,approx_grad = 1, factr=10., pgtol=1e-08, maxfun=float('Inf'))
    a = res[0]
    Sbg = SP.zeros_like(S[0])
    Kbg = SP.zeros_like(K[0])
    for i in range(len(S)):  Sbg+= a[i]*S[i]
    for i in range(len(K)):  Kbg+= a[i+len(S)]*K[i]

    return Sbg, Kbg


class CFastVDMM:
    """
    helper function for variance decomposition in limix
    This class mainly takes care of initialization and interpreation of results
    """
    
    """
    Methods:
        __init__(self,Y,T,F,C,F):    Constructor
        initialise:                  Initialise parameters
        fit:                         Fit phenos and returns the minimum with all the info
        fit_ntimes:                  Fit phenos ntimes with different random initialization and returns the minima order with respect to gp.LML
        getGP:                       Return the GP of the limix variacne decomposition class
        getParams:                   Returns the stack vector of the parameters of the trait covariances C_i
        getEmpTraitCov:              Returns the empirical trait covariance
        getEstTraitCov:              Returns the total trait covariance in the GP by summing up the trait covariances C_i
        getCovParams:                Calculates the inverse hessian of the -loglikelihood with respect to the parameters: this constitutes the covariance matrix of the posterior over paramters under laplace approximation
        estimateHeritabilities:      It fits the model with 1 global fixed effects and covariance matrix h1[p]*K+h2[p] for each trait p and returns the vectors h1 and h2
        exportMin:                   Export information about the min in the given h5py file or group
        exportMins:                  Export information about all the min in the given h5py file or group

    """
    
    def __init__(self,Y,C1,C2,K1,K2=None):
        """
        Args:
            Y:  NxP phenotype matrix
            C:  list of trait covariances
            K1: kernel matrix 1
            K2: kernel matrix 2
        """
        
        self.N=Y.shape[0]
        self.P=Y.shape[1]
        
        #Diagonalization of Part II
        if K2 == None:
            K2   = SP.eye(self.N)
            U2   = SP.eye(self.N)
            S2is = SP.eye(self.N)
            logdet_compl = 0
            K = K1
        else:
            eigen2, U2 = SP.linalg.eigh(K2)
            eigen2= SP.array(eigen2,dtype=float)
            S2is = SP.diag(1/SP.sqrt(eigen2))
            logdet_compl = self.P*SP.log(eigen2).sum()
            K = SP.dot(S2is,SP.dot(U2.T,SP.dot(K1,SP.dot(U2,S2is))))
                
        #Diagonalization of Part I
        eigen, U = SP.linalg.eigh(K)
        eigen = SP.array(eigen,dtype=float)
        U= SP.array(U,dtype=float)
        S   = SP.diag(eigen)
                
        # pheno transformation
        Yt = SP.zeros_like(Y.T)
        Upheno = SP.dot(U.T,SP.dot(S2is,U2.T))
        for p in range(self.P):
            Yt[p,:] = SP.dot(Upheno,Y[:,p])
        
        self.C1 = self.buildCF(C1)
        self.C2 = self.buildCF(C2)
        self.Yt = Yt
        self.eigen = eigen
        pass
    
    def buildCF(self,C):
        # Initialize CTraitCFs and wrap them in a Sum of matrices
        trait = SP.array(list(range(self.P)))[:,SP.newaxis]
        if type(C)!=list:
            Cout=C
            if Cout.getName()[0:2]=='CT':    Cout.setX(trait)
        else:
            Cout = limix.CSumCF()
            n_terms = len(C)
            for term_i in range(n_terms):
                if C[term_i].getName()[0:2]=='CT':    C[term_i].setX(trait)
                Cout.addCovariance(C[term_i])
        return Cout

    
    
    def initialise(self):
        """
        get random initialization of variances based on the empirical trait variance
        
        EmpVarY=self.getEmpTraitVar()
        temp=SP.rand(self.n_terms,self.P)
        N=temp.sum(0)
        temp=temp/N*EmpVarY
        for term_i in range(self.n_terms):
            if self.C[term_i].getName()=='CTDenseCF':
                params=(2*SP.rand(self.P)-1)*SP.sqrt(temp[term_i,:])
                self.C[term_i].setParams(params)
            elif self.C[term_i].getName()=='CTFixedCF':
                params=SP.array([SP.sqrt(temp[term_i,:].mean())])
                self.C[term_i].setParams(params)
            elif self.C[term_i].getName()=='CTDiagonalCF':
                params=SP.sqrt(temp[term_i,:])
                self.C[term_i].setParams(params)
            else:
                print 'Not implemented for %s' % self.C[term_i].getName()
                break
        """

    
    def fit(self,Params0=None,grad_threshold=1e-2):
        """
        fit a variance component model with the predefined design and the initialization and returns all the results
        """

        # GPVD initialization
        lik = limix.CLikNormalNULL()
        # Initial Params
        if Params0==None:
            n_params = self.C1.getNumberParams()
            n_params+= self.C2.getNumberParams()
            Params0 = SP.rand(n_params)
        # MultiGP Framework
        covar = []
        gp    = []
        mean  = []
        for i in range(self.N):
            covar.append(limix.CLinCombCF())
            covar[i].addCovariance(self.C1)
            covar[i].addCovariance(self.C2)
            coeff = SP.array([self.eigen[i],1])
            covar[i].setCoeff(coeff)
            mean.append(limix.CLinearMean(self.Yt[:,i],SP.eye(self.P)))
        gpVD = limix.CGPvarDecomp(covar[0],lik,mean[0],SP.ones(self.N),self.P,self.Yt,Params0)
        for i in range(self.N):
            gp.append(limix.CGPbase(covar[i],lik,mean[i]))
            gp[i].setY(self.Yt[:,i])
            gpVD.addGP(gp[i])
                
        # Optimization
        gpVD.initGPs()
        gpopt = limix.CGPopt(gpVD)
        LML0=-1.0*gpVD.LML()
        start_time = time.time()
        conv = gpopt.opt()
        time_train = time.time() - start_time
        LML=-1.0*gpVD.LML()
        LMLgrad = SP.linalg.norm(gpVD.LMLgrad()['covar'])
        Params = gpVD.getParams()['covar']
    
        # Check whether limix::CVarianceDecomposition.train() has converged
        if conv!=True or LMLgrad>grad_threshold or Params.max()>10:
            print('limix::CVarianceDecomposition::train has not converged')
            res=None
        else:
            res = {
                'Params0':          Params0,
                'Params':           Params,
                'LML':              SP.array([LML]),
                'LML0':             SP.array([LML0]),
                'LMLgrad':          SP.array([LMLgrad]),
                'time_train':       SP.array([time_train]),
                }
        return res
        pass
            

    
    
    def fit_ntimes(self,ntimes=10,grad_threshold=1e-2,dist_mins=1e-2):
        """
        fit phenos ntimes with different random initialization and returns the minima order with respect to gp.LML
        

        optima=[]
        LML=SP.zeros((1,0))
        
        for i in range(ntimes):
            
            print ".. Minimization %d" % i
            
            self.initialise()
            min=self.fit(grad_threshold=grad_threshold)
        
            if min!=None:
                temp=1
                for j in range(len(optima)):
                    if SP.linalg.norm(min['Params']-optima[j]['Params'])<dist_mins:
                        temp=0
                        optima[j]['counter']+=1
                        break
                if temp==1:
                    min['counter']=1
                    optima.append(min)
                    LML=SP.concatenate((LML,min['LML'][:,SP.newaxis]),1)
    
         # Order the list optima with respect to LML the first optimum has highest LML
        optima1=[]
        index = LML.argsort(axis = 1)[0,:][::-1]
        for i in range(len(optima)):
            optima1.append(optima[index[i]])
    
        return optima1
        """

    
    def getGP(self):
        """
        Returns the GP of the limix class CVarianceDecomposition
        
        vt = limix.CVarianceDecomposition()
        vt.setPheno(self.Y)
        vt.setFixed(self.F)
        for term_i in range(self.n_terms):
            vt.addCVTerm(self.C[term_i],self.K[term_i],self.T[term_i])
        vt.initGP()
        gp=vt.getGP()
        return gp
        """

    def getParams(self):
        """
        Returns the Parameters
        
        params=SP.concatenate([self.C[term_i].agetScales() for term_i in range(self.n_terms)])
        return params
        """

    def getEstTraitCov(self):
        """
        Returns the estimated trait covariance matrix
        
        TraitCovar=SP.zeros((self.P,self.P))
        for term_i in range(self.n_terms):
            TraitCovar+=self.C[term_i].getK0()
        return TraitCovar
        """

    def getCovParams(self,min):
        """
        USES LAPLACE APPROXIMATION TO CALCULATE THE COVARIANCE MATRIX OF THE OPTIMIZED PARAMETERS
        
        gp=min['gp']
        ParamMask=gp.getParamMask()['covar']
        std=SP.zeros(ParamMask.sum())
        H=gp.LMLhess(["covar"])
        It= (ParamMask[:,0]==1)
        H=H[It,:][:,It]
        Sigma = SP.linalg.inv(H)
        return Sigma
        """
    
    def getModelPosterior(self,min,Sigma=None):
        """
        USES LAPLACE APPROXIMATION TO CALCULATE THE BAYESIAN MODEL POSTERIOR
        
        if Sigma==None:
            Sigma = self.getCovParams(min)
        
        n_params = 0
        for term_i in range(self.n_terms):
            n_params += self.C[term_i].getNumberParams()
        
        ModCompl = 0.5*n_params*SP.log(2*SP.pi)+0.5*SP.log(SP.linalg.det(Sigma))
        
        RV = min['LML']+ModCompl
            
        return RV
        """
    

    def getEmpTraitCov(self):
        """
        Returns the empirical trait covariance matrix
        
        Y1=(self.Y).reshape((self.P,self.N))
        RV=SP.cov(Y1)
        return RV
        """
    
    def getEmpTraitVar(self):
        """
        Returns the vector of empirical trait variances
        
        if self.P==1:
            RV = self.getEmpTraitCov()
        else:
            RV = self.getEmpTraitCov().diagonal()
        return RV
        """




    def estimateHeritabilities(self,Kpop):
        """
        It fits the model with 1 fixed effects and covariance matrix h1[p]*K+h2[p] for each trait p and return the vectors h1 and h2
        
    
        h1 = SP.zeros(self.P)
        h2 = SP.zeros(self.P)
    
        for p in range(self.P):
            It = (self.T[0][:,0]==p)
            K= Kpop[It,:][:,It]
            N= K.shape[0]
            y=self.Y[It]
            h1[p], h2[p]=limix.CVarianceDecomposition.aestimateHeritability(y,SP.ones(N),K)
            if h1[p]<1e-6:   h1[p]=1e-6
            if h2[p]<1e-6:   h2[p]=1e-6
    
        return h1, h2
        """
    
    def exportMin(self,min,f,counter=0,laplace=0):
        """
        Export the min in the given h5py file or group    
        
        f.create_dataset("Params0",data=min["Params0"])
        f.create_dataset("LML0",data=min["LML0"])
        f.create_dataset("Params",data=min["Params"])
        f.create_dataset("LML",data=min["LML"])
        f.create_dataset("LMLgrad",data=min["LMLgrad"])
        f.create_dataset("time_train",data=min["time_train"])
        if counter!=0:
            f.create_dataset("counter",data=min["counter"])
        if laplace!=0:
            covParams = self.getCovParams(min)
            f.create_dataset("covParams",data=covParams)
            f.create_dataset("modelPosterior",data=self.getModelPosterior(min,Sigma=covParams))
        """
        
    
    def exportMins(self,mins,f,laplace=0):
        """
        Export all the min in the given h5py file or group  
        
        for min_i in range(len(mins)):
            g=f.create_group('min%d'%min_i)
            self.exportMin(mins[min_i],g,counter=1,laplace=laplace)
        """







# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.
import numpy as np
import scipy as sp
import glob
import pdb, sys, pickle
import matplotlib.pylab as plt
from optparse import OptionParser
# from util import *
import os, shutil
import subprocess as s
#import covar
#import one_chr
np.random.seed(1)
#from recode import *

os.putenv('FastLmmUseAnyMklLib', '1')
plink_path = 'plink' # 'D:/users/lippert/code/plink.exe'
fastlmm_path = 'fastlmmc' # 'D:/users/lippert/Projects/FaSTLMM/CPP/x64/MKL-Release/FastLmmC.exe'


def get_column(filename, col, skiprows = 1, dtype = float):
    import pdb
    f = open(filename, 'r')
    results = []
    skipped = 0
    for line in f:
        if skipped < skiprows:
            skipped += 1
            continue
        data = sp.array(line.strip('\n').split('\t'))
        results.append(data[col])

    try:
        results = np.asarray(results, dtype = dtype)
    except ValueError:
        results = np.asarray(results, dtype = str)
        results = np.asarray(results[results != 'NA'], dtype = dtype)
    return results

#1) determine covariance matrix
#1.1) full RRM
#1.2) SNPs list
#1.2.1) subsampling (equally spaced)
#1.2.2) select
#1.2.2.1) linear regression - lambda
#1.2.2.2) linear regression - out of sample
#2) run fastlmm
#3) plot results

def load_pvals(filename, return_info = False, linreg = False):
    if linreg:
        i_pval = 4
    else:
        i_pval = 5
    if return_info:
        index = sp.array([i_pval, 1, 3])
    else:
        index = i_pval
    pvals = get_column(filename,index)
    return pvals

def extract_topN(N, selsnps_file, linreg_out):
    ordered_snps = get_column(linreg_out, 0, dtype = str)[:N]
    np.savetxt(selsnps_file, ordered_snps, fmt = '%s')

def run_select_topN(pheno_file,linreg_out,file_test=None,bfile_test = None,tfile_test = None,bfile_sim = None,file_sim = None,tfile_sim= None, out_dir='./tmp', covariates = None, excl_dist = None, excl_pos =
None, chr_only = None, quiet = False, refit_delta = False, command=None, Nmin=0, Nmax=1000, increment=100 ,Nsnps = None,recompute = False):
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    selsnps_file =  os.path.join(out_dir, 'selectlist.txt')
    extract_topN( Nmax,selsnps_file, linreg_out )

    if Nsnps is None:
        Nsnps = sp.arange(Nmin,Nmax,increment)
    lambdas = sp.zeros(Nsnps.shape)
    for i in range(Nsnps.shape[0]):
        current_dir = 'N%d'% Nsnps[i]
        out_dir_current = os.path.join(out_dir, current_dir)
        if Nsnps[i]>0:
            run_fastlmm(pheno_file,file_test,bfile_test,tfile_test,bfile_sim,file_sim,tfile_sim, out_dir_current, covariates, excl_dist, excl_pos, chr_only, quiet, Nsnps[i],selsnps_file,refit_delta, command, True,recompute=recompute)
            out_file_current = os.path.join(out_dir_current, 'results_LMM.txt')
            pvals = get_column(out_file_current,5)
        else:
            run_linreg(pheno_file,file_test,bfile_test,tfile_test, out_dir_current, covariates, chr_only, quiet, command, True,recompute=recompute)
            out_file_current = os.path.join(out_dir_current, 'results_linreg.txt')
            pvals = get_column(out_file_current,4)
        lambdas[i] = estimate_lambda(pvals.flatten())
    return Nsnps, lambdas

def run_linreg(pheno_file,file_test=None,bfile_test = None,tfile_test = None, out_dir='./tmp', covariates = None, chr_only = None, quiet = False, command=None, run = True,recompute = False,**kw_args):
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    out_file = os.path.join(out_dir, 'results_linreg.txt')

    if command is None:
        command = ('%s -pheno %s -verboseOut -out %s -linreg' %(fastlmm_path, pheno_file, out_file))

    #SNP files for testing
    if bfile_test is not None: #binary PED format
        command += ' -bfile %s' % bfile_test
    elif tfile_test is not None: #TPED format
        command += ' -tfile %s' % tfile_test
    elif file_test is not None: #PED format
        command += ' -file %s' % file_test
    else:
        pass

    if covariates != None:
        command += " -covar %s" % covariates

    out_log = os.path.join(out_dir,'results_linreg.log')

    if len(glob.glob(out_log)) == 1:
        f = open(out_log,'r')
        for line in f:
            commandold = line.strip('\n')
            if commandold == command:
                run  = False
    np.savetxt(out_log, sp.array([command]),'%s')
    if (run or recompute):
        if quiet:
            print('Running linear regression ...')
            print(('%s' % (command)))
            s.check_call(command, shell = True, stdout = fnull, stderr = fnull)
        else:
            s.check_call(command, shell = True)
    else:
        print('returning without running linear regression')
    return command

def run_fastlmm(pheno_file,file_test=None,bfile_test = None,tfile_test = None,bfile_sim = None,file_sim = None,tfile_sim= None, out_dir='./tmp', covariates = None, excl_dist = None, excl_pos = None, chr_only = None, quiet = False, N=None,selsnps_file = None,refit_delta = False, command=None,run = True,recompute=False,**kw_args):
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    out_file = os.path.join(out_dir, 'results_LMM.txt')
    if command is None:
        command = ('%s -pheno %s -verboseOut -out %s' %(fastlmm_path, pheno_file, out_file))
    #SNP files for testing
    if bfile_test is not None: #binary PED format
        command += ' -bfile %s' % bfile_test
    elif tfile_test is not None: #TPED format
        command += ' -tfile %s' % tfile_test
    elif file_test is not None: #PED format
        command += ' -file %s' % file_test
    else:
        pass
    #SNP files for RRM
    if bfile_sim is not None: #binary PED format
        command += ' -bfilesim %s' % bfile_sim
    elif tfile_sim is not None: #TPED format
        command += ' -tfilesim %s' % tfile_sim
    elif file_sim is not None: #PED format
        command += ' -filesim %s' % file_sim
    else:
        pass
    if covariates != None:
        command += " -covar %s" % covariates

    if N is not None:
        command+= " -extractSimTopK %s %d" %(selsnps_file,N)

    if excl_dist != None:
        command += ' -excludebygeneticdistance %f' % (excl_dist)
    elif excl_pos != None:
        command += ' -excludebyposition %d' %  (excl_pos)
    out_log = os.path.join(out_dir,'results_LMM.log')

    if len(glob.glob(out_log)) == 1:
        f = open(out_log,'r')
        for line in f:
            commandold = line.strip('\n')
            if commandold == command:
                run  = False
    np.savetxt(out_log, sp.array([command]),'%s')
    if (run or recompute):
        if quiet:
            print('Running LMM ...')
            print(command)
            s.check_call(command, shell = True, stdout = fnull, stderr = fnull)
        else:
            s.check_call(command, shell = True)
    else:
        print('returning without running LMM')
    return command



def get_options(opt = None):
    usage = 'usage: %prog [options] snps_file pheno_file'
    parser = OptionParser(usage = usage)
    parser.add_option('-f', '--file', dest = 'file_test', action = 'store', type = 'string', help = 'Plink PED fileset', default = None)
    parser.add_option('-b', '--bfile', dest = 'bfile_test', action = 'store', type = 'string', help = 'Plink binary fileset', default = None)
    parser.add_option('-t', '--tfile', dest = 'tfile_test', action = 'store', type = 'string', help = 'Plink TPED fileset', default = None)
    parser.add_option('-K', '--filesim', dest = 'file_sim', action = 'store', type = 'string', help = 'Plink PED fileset', default = None)
    parser.add_option('-L', '--bfilesim', dest = 'bfile_sim', action = 'store', type = 'string', help = 'Plink binary fileset', default = None)
    parser.add_option('-Q', '--tfilesim', dest = 'tfile_sim', action = 'store', type = 'string', help = 'Plink TPED fileset', default = None)
    parser.add_option('-N', '--n-snps', dest = 'N', action = 'store', type = 'int', help = 'number of SNPs to use to estimate Kinship (realized relationship) matrix (default all)', default = None)
    parser.add_option('-o', '--out', dest = 'out_dir', action = 'store', type = 'string', help = 'Output directory (default ./tmp)', default = './tmp')
    parser.add_option('-d', '--delimiter', dest = 'delimiter', action = 'store', type = 'string', help = 'delimiter between fields of the phenotype file', default = '\t')
    parser.add_option('-m', '--pheno-missing-value', dest = 'missing', action = 'store', type = 'float', help = 'missing value for the phenotype file (default -9)', default = -9.0)
    parser.add_option('-q', '--quiet', dest = 'quiet', action = 'store_true', help = 'show output from calls', default = False)
    parser.add_option('-C', '--covariates_file', dest = 'covariates', action = 'store', type = 'string', help = 'covariates', default = None)
    parser.add_option('-D', '--exclude_distance', dest = 'excl_dist', action = 'store', type = 'float', help = 'exclude by distance (in cM)', default = None)
    parser.add_option('-P', '--exclude_position', dest = 'excl_pos', action = 'store', type = 'int', help = 'exclude by position (in bp)', default = None)
    parser.add_option('-s', '--spaced', dest = 'spaced', action = 'store_true',  help = 'equally spaced regressors')
    parser.add_option('-T', '--testing', dest = 'testing', action = 'store_true',  help = 'only run the final LMM')
    parser.add_option('-R', '--one_chr', dest = 'chr_only', action = 'store_true',  help = 'test on one chr and build kernel on the rest')
    parser.add_option('-p', '--pheno', dest = 'pheno_file', action = 'store',  type = 'string', help = 'phenotype file')
    if opt is not None:
        (options, args) = parser.parse_args(opt)
    else:
        (options, args) = parser.parse_args()
    return options, args

if __name__ == '__main__':
    (options, args) = get_options()

    #command_LMM = run_fastlmm(pheno_file = options.pheno_file,file_test=options.file_test,bfile_test = options.bfile_test,tfile_test = options.tfile_test,bfile_sim = options.bfile_sim,file_sim = options.file_sim,tfile_sim= options.tfile_sim, out_dir=options.out_dir, covariates = options.covariates, excl_dist = options.excl_dist, excl_pos = options.excl_pos, chr_only = options.chr_only, testing = False, quiet = False, N=None,refit_delta = False)

#    command_linreg = run_linreg(pheno_file = options.pheno_file, file_test=options.file_test, bfile_test = options.bfile_test, tfile_test = options.tfile_test, out_dir=options.out_dir, covariates = options.covariates, chr_only = options.chr_only, testing = False, quiet = False)

    linreg_out = os.path.join(options.out_dir, 'results_linreg.txt')
    pvals = load_pvals(linreg_out,True,True )



    #Nsnps0,lambdas0 = run_select_topN(pheno_file= options.pheno_file, linreg_out=linreg_out,file_test=options.file_test,bfile_test = options.bfile_test,tfile_test = options.tfile_test,bfile_sim = options.bfile_sim,file_sim = options.file_sim, tfile_sim = options.tfile_sim, out_dir=options.out_dir, covariates = options.covariates, excl_dist = options.excl_dist, excl_pos = options.excl_pos, chr_only = options.chr_only, quiet = False, refit_delta = False, command=None, Nmin=0, Nmax=100, increment=10 )

    #Nsnps = sp.arange(20)

    #Nsnps,lambdas = run_select_topN(pheno_file= options.pheno_file, linreg_out=linreg_out,file_test=options.file_test,bfile_test = options.bfile_test,tfile_test = options.tfile_test,bfile_sim = options.bfile_sim,file_sim = options.file_sim, tfile_sim = options.tfile_sim, out_dir=options.out_dir, covariates = options.covariates, excl_dist = options.excl_dist, excl_pos = options.excl_pos, chr_only = options.chr_only, quiet = False, refit_delta = False, command=None, Nmin=0, Nmax=100, increment=10 ,Nsnps = Nsnps)



'''
Created on Sep 19, 2013

@author: johannes
'''
# create some test cases
import scipy as SP
from limix.ensemble import lmm_forest_utils as utils
import h5py
from limix.ensemble.lmm_forest import Forest as MF
import os
import unittest


class TestMixedForest(unittest.TestCase):

    def setUp(self, n=100, m=1):
        self.dir_name = os.path.dirname(os.path.realpath(__file__))
        self.data = h5py.File(os.path.join(self.dir_name,
                                           'test_data/lmm_forest_toy_data.h5'),
                              'r')
        SP.random.seed(1)
        self.x, self.y = utils.lin_data_cont_predictors(n=n,m=m)
        self.n, self.m = self.x.shape
        [self.train, self.test] = utils.crossValidationScheme(2,self.n)
        self.n_estimators = 100

    @unittest.skip("someone has to fix it")
    def test_toy_data_rand(self):
        y_conf = self.data['y_conf'].value
        kernel = self.data['kernel'].value
        X = self.data['X'].value
        # This is a non-random cross validation
        (training, test) = utils.crossValidationScheme(2, y_conf.size)
        lm_forest = MF(kernel=kernel[SP.ix_(training, training)],
                       sampsize=.5, verbose=0, n_estimators=100)
        lm_forest.fit(X[training], y_conf[training])
        response_tot = lm_forest.predict(X[test],
                                         kernel[SP.ix_(test, training)])
        random_forest = MF(kernel='iid')
        random_forest.fit(X[training], y_conf[training])
        response_iid = random_forest.predict(X[test])
        response_fixed = lm_forest.predict(X[test])
        feature_scores_lmf = lm_forest.log_importance
        feature_scores_rf = random_forest.log_importance
        # All consistency checks
        err = (feature_scores_lmf-self.data['feature_scores_lmf'].value).sum()
        self.assertTrue(SP.absolute(err) < 10)
        err = (feature_scores_rf-self.data['feature_scores_rf'].value).sum()
        self.assertTrue(SP.absolute(err) < 10)
        err = SP.absolute(self.data['response_tot'] - response_tot).sum()
        self.assertTrue(SP.absolute(err) < 2)
        err = SP.absolute(self.data['response_fixed'] - response_fixed).sum()
        self.assertTrue(SP.absolute(err) < 4)
        err = SP.absolute(self.data['response_iid'] - response_iid).sum()
        self.assertTrue(SP.absolute(err) < 8)

    def test_delta_updating(self):
        n_sample = 100
        # A 20 x 2 random integer matrix
        X = SP.empty((n_sample, 2))
        X[:, 0] = SP.arange(0, 1, 1.0/n_sample)
        X[:, 1] = SP.random.rand(n_sample)
        sd_noise = .5
        sd_conf = .5
        noise = SP.random.randn(n_sample, 1)*sd_noise

        # print 'true delta equals', (sd_noise**2)/(sd_conf**2)
        # Here, the observed y is just a linear function of the first column
        # in X and # a little independent gaussian noise
        y_fixed = (X[:, 0:1] > .5)*1.0
        y_fn = y_fixed + noise

        # Divide into training and test sample using 2/3 of data for training
        training_sample = SP.zeros(n_sample, dtype='bool')
        training_sample[
            SP.random.permutation(n_sample)[:SP.int_(.66*n_sample)]] = True
        test_sample = ~training_sample

        kernel = utils.getQuadraticKernel(X[:, 0], d=0.0025) +\
            1e-3*SP.eye(n_sample)
        # The confounded version of y_lin is computed as
        y_conf = sd_conf*SP.random.multivariate_normal(SP.zeros(n_sample),
                                                       kernel, 1).reshape(-1, 1)
        y_tot = y_fn + y_conf
        # Selects rows and columns
        kernel_train = kernel[SP.ix_(training_sample, training_sample)]
        kernel_test = kernel[SP.ix_(test_sample, training_sample)]
        lm_forest = MF(kernel=kernel_train, update_delta=False, max_depth=1,
                       verbose=0)
        # Returns prediction for random effect
        lm_forest.fit(X[training_sample], y_tot[training_sample])
        response_lmf = lm_forest.predict(X[test_sample], k=kernel_test)

        # print 'fitting forest (delta-update)'
        # earn random forest, not accounting for the confounding
        random_forest = MF(kernel=kernel_train, update_delta=True, max_depth=5,
                           verbose=0)
        random_forest.fit(X[training_sample], y_tot[training_sample])
        response_rf = random_forest.predict(X[test_sample], k=kernel_test)

    def test_kernel_builing(self):
        X = (SP.random.rand(5, 10) > .5)*1.0
        kernel = utils.estimateKernel(X, scale=False)
        small_kernel = utils.estimateKernel(X[:, 0:5], scale=False)
        small_kernel_test = utils.update_Kernel(kernel, X[:, 5:], scale=False)
        self.assertAlmostEqual((small_kernel -
                                small_kernel_test).sum(), 0)

    def test_depth_building(self):
        self.setUp(m=10)
        X = self.x.copy()
        X -= X.mean(axis=0)
        X /= X.std(axis=0)
        kernel = SP.dot(X, X.T)
        train = SP.where(self.train)[0]
        test = SP.where(~self.train)[0]
        model = MF(fit_optimal_depth=True, max_depth=3,
                   kernel=kernel[SP.ix_(train, train)])
        model.fit(self.x[self.train], self.y[self.train],
                  fit_optimal_depth=True)
        prediction_1 = model.predict(X[test], k=kernel[test, train],
                                     depth=model.opt_depth)
        # Grow to end
        model.further()
        # Prediction again
        prediction_2 = model.predict(X[test], k=kernel[test, train],
                                     depth=model.opt_depth)
        self.assertEqual((prediction_1 - prediction_2).sum(), 0.0)

    @unittest.skip("someone has to fix it")
    def test_forest_stump_recycling(self):
        self.setUp(m=5)
        SP.random.seed(42)
        model = MF(fit_optimal_depth=True, kernel='iid',
                   build_to_opt_depth=True)
        model.fit(self.x[self.train], self.y[self.train])
        prediction_1 = model.predict(self.x[self.test], depth=model.opt_depth)
        model.fit(self.x[self.train], self.y[self.train], recycle=True)
        prediction_2 = model.predict(self.x[self.test], depth=model.opt_depth)
        self.assertGreater(.7, ((prediction_1 - prediction_2)**2).sum())

    @unittest.skip("someone has to fix it")
    def test_normalization_kernel(self):
        #SP.random.seed(42)
        n = 50
        m = 100
        X = (SP.random.rand(n, m) > .5)*1.
        X_test = (SP.random.rand(10, m) > .5)*1.
        K = utils.estimateKernel(X)
        y = SP.random.rand(n, 1)
        SP.random.seed(1)
        mf = MF(kernel=K)
        mf.fit(X, y)
        results_1 = mf.predict(X_test)

        X -= X.mean(axis=0)
        X /= X.std(axis=0)
        X_test -= X_test.mean(axis=0)
        X_test /= X_test.std(axis=0)

        SP.random.seed(1)
        mf = MF(kernel=K)
        mf.fit(X, y)
        results_2 = mf.predict(X_test)
        self.assertEqual(results_1.sum(), results_2.sum())

    def polynom(self, x):
        return -x + x**3

    def complete_sample(self, x, mean=0, var=.3**2):
        return self.polynom(x) + SP.random.randn(x.size) * SP.sqrt(var) + mean

    def test_covariate_shift(self):
        n_sample = 100
        # Biased training
        var_bias = .5**2
        mean_bias = .7
        x_train = SP.random.randn(n_sample)*SP.sqrt(var_bias) + mean_bias
        y_train = self.complete_sample(x_train)

        # Unbiased test set
        var = .3**2
        mean = 0

        x_test = SP.random.randn(n_sample)*SP.sqrt(var) + mean
        x_complete = SP.hstack((x_train, x_test))

        kernel = utils.getQuadraticKernel(x_complete, d=1) +\
            10 * SP.dot(x_complete.reshape(-1, 1), x_complete.reshape(1, -1))
        kernel = utils.scale_K(kernel)
        kernel_train = kernel[SP.ix_(SP.arange(x_train.size),
                                     SP.arange(x_train.size))]
        kernel_test = kernel[SP.ix_(SP.arange(x_train.size, x_complete.size),
                             SP.arange(x_train.size))]

        mf = MF(n_estimators=100, kernel=kernel_train, min_depth=0,
                subsampling=False)
        mf.fit(x_train.reshape(-1, 1), y_train.reshape(-1, 1))
        response_gp = mf.predict(x_test.reshape(-1, 1), kernel_test, depth=0)
        self.assertTrue(((response_gp - self.polynom(x_test))**2).sum() < 2.4)

if __name__ == '__main__':
    unittest.main()



import sys
import numpy as np
from .covar_base import Covariance
from .freeform import FreeFormCov
from .diagonal import DiagonalCov 
from hcache import cached
from limix.core.type.exception import TooExpensiveOperationError
from limix.core.utils import my_name
from .util import msg_too_expensive_dim
from limix.utils.svd_utils import svd_reduce
import scipy as sp
import scipy.linalg as la
import numpy.linalg as nla
import warnings

import pdb

_MAX_DIM = 5000

class CategoricalLR(Covariance):

    def __init__(self, Cr=None, G=None, Ie=None):
        """
        Args:
            Cr:     Limix covariance matrix for Cr (dimension 2) 
                        must be freeform, lowrank or block
            G:      [dim, rank] numpy covariance matrix for G
            Ie:     environment labels Ie \in {0,1}^dim
        """
        Covariance.__init__(self)
        self._Cr_act = True
        self._Cn_act = True
        self.setColCovars(Cr)
        self.G = G
        self.Ie = Ie
        self._use_to_predict = False

    #####################
    # Properties
    #####################
    @property
    def G(self):
        return self._G

    @property
    def Cr(self):
        return self._Cr

    @property
    def Cn(self):
        return self._Cn

    @property
    def dim(self):
        return self._dim

    #####################
    # Setters
    #####################
    @G.setter
    def G(self,value):
        assert value is not None, 'G cannot be set to None.'
        self._dim = value.shape[0]
        # perform svd on G
        # excludes eigh < 1e-8 and recalcs G
        _value, U, S, V = svd_reduce(value)
        self._G  = _value
        self._Ug = U
        self._Sg = S
        self._Vg = V
        self.clear_all()

    # normal setter for col covars
    def setColCovars(self, Cr = None, rank = 1):
        assert Cr is not None, 'Cr has to be specified.'
        self._Cr = Cr 
        self._Cn = DiagonalCov(2)
        # register
        self._Cr.register(self.clear_all)
        self._Cn.register(self.clear_all)
        self.clear_all()

    #####################
    # Activation handling
    #####################
    @property
    def act_Cr(self):
        return self._Cr_act

    @act_Cr.setter
    def act_Cr(self, act):
        self._Cr_act = bool(act)
        self._notify()

    @property
    def act_Cn(self):
        return self._Cn_act

    @act_Cn.setter
    def act_Cn(self, act):
        self._Cn_act = bool(act)
        self._notify()

    def _actindex2index(self, i):
        nCr = self.Cr.getNumberParams()
        i += nCr * int(not self._Cr_act)
        return i

    def _index2actindex(self, i):
        nCr = self.Cr.getNumberParams()
        i -= nCr * int(not self._Cr_act)
        return i


    #####################
    # Params handling
    #####################
    def setParams(self, params):
        nCr = int(self._Cr_act) * self.Cr.getNumberParams()
        nCn = int(self._Cn_act) * self.Cn.getNumberParams()

        if len(params) != self.getNumberParams():
            raise ValueError("The number of parameters passed to setParams "
                             "differs from the number of active parameters.")
        if self._Cr_act:
            self.Cr.setParams(params[:nCr])
        if self._Cn_act:
            self.Cn.setParams(params[nCr:])

    def getParams(self):
        params = []
        if self._Cr_act:
            params.append(self.Cr.getParams())
        if self._Cn_act:
            params.append(self.Cn.getParams())
        if len(params) == 0:
            return np.array([])
        return sp.concatenate(params)

    def getNumberParams(self):
        return (int(self._Cr_act) * self.Cr.getNumberParams() +
                int(self._Cn_act) * self.Cn.getNumberParams())


    #####################
    # Non-cached methods
    #####################
    def solve(self, M):
        DiM = self.d_inv()[:, sp.newaxis] * M
        DiWHiWDiM = sp.dot(self.DiW(), sp.dot(self.H_inv(), sp.dot(self.W().T, DiM)))
        RV = DiM - DiWHiWDiM
        return RV

    def K_grad_i_dot(self, M, i):
        if i < self.Cr.getNumberParams():
            R = sp.dot(self.W(), sp.dot(self.W_grad_i(i).T, M))
            R+= sp.dot(self.W_grad_i(i), sp.dot(self.W().T, M))
        else:
            R = self.d_grad_i(i-self.Cr.getNumberParams())[:, sp.newaxis] * M
        return R

    def Kh_dot(self, M):
        R = (self.d_inv()**(-0.5))[:,sp.newaxis]*M
        R+= sp.dot(self.W(), sp.dot(self.W().T, M))
        return R

    #####################
    # Cached
    #####################
    @cached('covar_base')
    def W(self):
        RV = sp.zeros((self.dim, self.Cr.X.shape[1]*self.G.shape[1]))
        RV[self.Ie] = sp.kron(self.Cr.X[0], self.G[self.Ie])
        RV[~self.Ie] = sp.kron(self.Cr.X[1], self.G[~self.Ie])
        return RV 

    @cached('covar_base')
    def d_inv(self):
        RV = sp.zeros(self.dim)
        RV[self.Ie] = 1. / self.Cn.variance[0]
        RV[~self.Ie] = 1. / self.Cn.variance[1]
        return RV 

    @cached('covar_base')
    def DiW(self):
        return self.d_inv()[:, sp.newaxis] * self.W()

    @cached('covar_base')
    def WDiW(self):
        return sp.dot(self.W().T, self.DiW())

    @cached('covar_base')
    def H_chol(self):
        return la.cholesky(self.WDiW()+sp.eye(self.WDiW().shape[0])).T

    @cached('covar_base')
    def H_inv(self):
        return la.cho_solve((self.H_chol(), True), sp.eye(self.W().shape[1]))

    @cached('covar_base')
    def W_grad_i(self, i):
        RV = sp.zeros((self.dim, self.Cr.X.shape[1]*self.G.shape[1]))
        RV[self.Ie] = sp.kron(self.Cr.Xgrad(i)[0], self.G[self.Ie])
        RV[~self.Ie] = sp.kron(self.Cr.Xgrad(i)[1], self.G[~self.Ie])
        return RV 

    @cached('covar_base')
    def d_grad_i(self, i):
        RV = sp.zeros(self.dim)
        if i==0:
            RV[self.Ie] = self.Cn.variance[0]
        elif i==1:
            RV[~self.Ie] = self.Cn.variance[1]
        return RV 

    #####################
    # Overwritten covar_base methods
    #####################
    @cached('covar_base')
    def K(self):
        print('DO NOT USE ME')
        return sp.dot(self.W(), self.W().T) + sp.diag(self.d_inv()**(-1))

    @cached('covar_base')
    def logdet(self):
        return 2*sp.log(sp.diag(self.H_chol())).sum() - sp.log(self.d_inv()).sum()

    @cached('covar_base')
    def logdet_grad_i(self,i):
        if i >= self.getNumberParams():
            raise ValueError("Trying to retrieve the gradient over a "
                             "parameter that is inactive.")
        i = self._actindex2index(i)
        if i < self.Cr.getNumberParams():
            diag = 2*(self.W()*self.W_grad_i(i)).sum(1)
            #WDiW_grad_i = sp.dot(self.DiW().T, self.W_grad_i(i))
            #WDiW_grad_i+= WDiW_grad_i.T 
            #WDidKDiW = sp.dot(WDiW_grad_i, self.WDiW())
            _WDidKDiW = sp.dot(sp.dot(self.DiW().T, self.W_grad_i(i)), self.WDiW())
            WDidKDiW = _WDidKDiW + _WDidKDiW.T 
        else:
            diag = self.d_grad_i(i-self.Cr.getNumberParams())
            WDidKDiW = sp.dot(self.DiW().T, diag[:, sp.newaxis]*self.DiW())
        rv = sp.dot(self.d_inv(), diag)
        rv-= (self.H_inv() * WDidKDiW).sum()
        return rv

    @cached('covar_base')
    def K_grad_i(self,i):
        print('DO NOT USE ME')
        if i >= self.getNumberParams():
            raise ValueError("Trying to retrieve the gradient over a "
                             "parameter that is inactive.")
        i = self._actindex2index(i)
        if i < self.Cr.getNumberParams():
            _R = sp.dot(self.W(), self.W_grad_i(i).T)
            R = _R + _R.T
        else:
            R = sp.diag(self.d_grad_i(i-self.Cr.getNumberParams()))
        return R

if __name__=='__main__':

    import pylab as pl
    pl.ion()

    N = 100; S = 10
    G = 1. * (sp.rand(N, S)<0.2)
    G-= G.mean(0); G/= G.std(0); G /= sp.sqrt(S)
    Ie = sp.rand(N)<0.5
    Cr = FreeFormCov(2, jitter=0.)
    C = CategoricalLR(Cr, G, Ie)
    C.setRandomParams()

    # calc WW using W
    WW = sp.dot(C.W(), C.W().T)
    # calc WW directly
    Cr_big = sp.zeros((N,N))
    Cr_big[sp.ix_(Ie, Ie)]  = Cr.K()[0,0]
    Cr_big[sp.ix_(Ie, ~Ie)] = Cr.K()[0,1]
    Cr_big[sp.ix_(~Ie, Ie)] = Cr.K()[1,0]
    Cr_big[sp.ix_(~Ie, ~Ie)]= Cr.K()[1,1]
    WW1 = Cr_big * sp.dot(G, G.T)
    print(('WW:', ((WW1-WW)**2).mean()))

    pdb.set_trace()
    
    for i in range(3):
        # calc D(WW)
        _WW_grad = sp.dot(C.W_grad_i(i), C.W().T)
        WW_grad = _WW_grad + _WW_grad.T
        # calc WW directly
        Cr_big = sp.zeros((N,N))
        Cr_big[sp.ix_(Ie, Ie)]  = Cr.K_grad_i(i)[0,0]
        Cr_big[sp.ix_(Ie, ~Ie)] = Cr.K_grad_i(i)[0,1]
        Cr_big[sp.ix_(~Ie, Ie)] = Cr.K_grad_i(i)[1,0]
        Cr_big[sp.ix_(~Ie, ~Ie)]= Cr.K_grad_i(i)[1,1]
        WW1_grad = Cr_big * sp.dot(G, G.T)
        print(('WWgrad %d:' % i, ((WW1_grad-WW_grad)**2).mean()))

    # inverse
    y = sp.randn(N, 1)
    Kiy1 = C.solve(y)
    Kiy2 = sp.dot(C.inv(), y)
    print((((Kiy1-Kiy2)**2).mean()))

    # logdet
    ld1 = C.logdet()
    ld2 = sp.log(C.S()).sum() 
    print((ld1-ld2))

    # logdet_grad_i
    for i in range(C.getNumberParams()):
        ldg1 = C.logdet_grad_i(i)
        ldg2 = sp.dot(C.inv(), C.K_grad_i(i)).diagonal().sum()
        print(('logdet grad %d:' % i, ldg1-ldg2))

    # K_grad_dot
    M = sp.randn(N, 1)
    for i in range(C.getNumberParams()):
        DKM = C.K_grad_i_dot(M, i) 
        DKM1 = sp.dot(C.K_grad_i(i), M)
        print(('Kgrad%d.dot:' % i, ((DKM-DKM1)**2).mean()))

    pdb.set_trace()




import pdb
import scipy as SP

class ALik(object):
    """abstract class for arbitrary likelihood model"""
    pass

class GaussLikISO(ALik):
    """Gaussian isotropic likelihood model
    """
    def K(self,theta,n_dimensions):
        sigma = SP.exp(2*theta[0])
        Knoise = sigma*SP.eye(n_dimensions)
        return Knoise

    def Kdiag(self,theta,n_dimensions):
        sigma = SP.exp(2*theta[0])
        return sigma*SP.ones(n_dimensions)

    def Kgrad_theta(self,theta,n_dimensions,i):
        assert i==0, 'unknown hyperparameter'
        K = self.K(theta,n_dimensions)
        return 2*K

    def Kdiag_grad_theta(self,theta,n_dimensions,i):
        assert i==0, 'unknown hyperparameter'
        Kdiag = self.Kdiag(theta,n_dimensions)
        return 2*Kdiag




# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.

"""
PANAMA module in limix
"""

import limix.deprecated
import limix.deprecated.stats.fdr as fdr
from limix.deprecated.stats.pca import *
from . import qtl
import scipy as sp 
import pdb
import scipy.linalg as linalg
import time

class PANAMA:
    """PANAMA class"""
    def __init__(self,data=None,X=None,Y=None,Kpop=None,use_Kpop=True,standardize=True):
        """
        Args:
            data: data object to feed form
            X: alternatively SNP data
            Y: alternativel expression data 
            Kpop: Kpop
            use_Kpop: if True (default), Kpop is considered in the model
            standardize: if True, phenotypes are standardized
        """
        self.use_Kpop = use_Kpop

        if data is not None:
            self.Kpop = data.getCovariance()
            self.Y = data.getPhenotypes()[0].copy()
            self.X = data.gtGenotypes()
        else:
            self.X = X
            self.Y = Y
        if Kpop is not None:
            self.Kpop = Kpop
        elif self.X is not None:
            self.Kpop = sp.dot(self.X,self.X.T)
            self.Kpop /= self.Kpop.diagonal().mean()
        else:
            assert use_Kpop==False, 'no Kpop'
        
        if standardize:
            self.Y -= self.Y.mean(axis=0)
            self.Y /= self.Y.std(axis=0)
        self.N = self.Y.shape[0]
        self.P = self.Y.shape[1]
        if self.X is not None:
            self.S = self.X.shape[1]
            assert self.X.shape[0]==self.Y.shape[0], 'data size missmatch'
        pass

    def train(self,rank=20,Kpop=True,LinearARD=False):
        """train panama module"""

        if 0:
            covar  = limix.deprecarted.CCovLinearISO(rank)
            ll  = limix.deprecated.CLikNormalIso()
            X0 = sp.random.randn(self.N,rank)
            X0 = PCA(self.Y,rank)[0]
            X0 /= sp.sqrt(rank)
            covar_params = sp.array([1.0])
            lik_params = sp.array([1.0])

            hyperparams = limix.deprecated.CGPHyperParams()
            hyperparams['covar'] = covar_params
            hyperparams['lik'] = lik_params
            hyperparams['X']   = X0
        
            constrainU = limix.deprecated.CGPHyperParams()
            constrainL = limix.deprecated.CGPHyperParams()
            constrainU['covar'] = +5*sp.ones_like(covar_params);
            constrainL['covar'] = 0*sp.ones_like(covar_params);
            constrainU['lik'] = +5*sp.ones_like(lik_params);
            constrainL['lik'] = 0*sp.ones_like(lik_params);

        if 1:
            covar  = limix.deprecated.CSumCF()
            if LinearARD:
                covar_1 =  limix.deprecated.CCovLinearARD(rank)
                covar_params = []
                for d in range(rank):
                    covar_params.append(1/sp.sqrt(d+2))
            else:
                covar_1 =  limix.deprecated.CCovLinearISO(rank)
                covar_params = [1.0]
            covar.addCovariance(covar_1)

            if self.use_Kpop:
                covar_2 =  limix.deprecated.CFixedCF(self.Kpop)
                covar.addCovariance(covar_2)
                covar_params.append(1.0)

            ll  = limix.deprecated.CLikNormalIso()
            X0 = PCA(self.Y,rank)[0]
            X0 /= sp.sqrt(rank)
            covar_params = sp.array(covar_params)
            lik_params = sp.array([1.0])

            hyperparams = limix.deprecated.CGPHyperParams()
            hyperparams['covar'] = covar_params
            hyperparams['lik'] = lik_params
            hyperparams['X']   = X0
        
            constrainU = limix.deprecated.CGPHyperParams()
            constrainL = limix.deprecated.CGPHyperParams()
            constrainU['covar'] = +5*sp.ones_like(covar_params);
            constrainL['covar'] = -5*sp.ones_like(covar_params);
            constrainU['lik'] = +5*sp.ones_like(lik_params);

            
        gp=limix.deprecated.CGPbase(covar,ll)
        gp.setY(self.Y)
        gp.setX(X0)
        lml0 = gp.LML(hyperparams)
        dlml0 = gp.LMLgrad(hyperparams)        
        gpopt = limix.deprecated.CGPopt(gp)
        gpopt.setOptBoundLower(constrainL);
        gpopt.setOptBoundUpper(constrainU);

        t1 = time.time()
        gpopt.opt()
        t2 = time.time()

        #Kpanama
        self.Xpanama = covar_1.getX()
        if LinearARD:
            self.Xpanama /= self.Xpanama.std(0)
        self.Kpanama = covar_1.K()
        self.Kpanama/= self.Kpanama.diagonal().mean()

        # Ktot
        self.Ktot = covar_1.K()
        if self.use_Kpop:
            self.Ktot += covar_2.K()
        self.Ktot/= self.Ktot.diagonal().mean()

        #store variances
        V = {}
        if LinearARD:
            V['LinearARD'] = covar_1.getParams()**2*covar_1.getX().var(0)
        else:
            V['Kpanama'] = sp.array([covar_1.K().diagonal().mean()])
        if self.use_Kpop:
            V['Kpop'] = sp.array([covar_2.K().diagonal().mean()])
        V['noise'] = gp.getParams()['lik']**2
        self.varianceComps = V

    def get_Xpanama(self):
        """
        Returns:
            matrix of Xs
        """
        return self.Xpanama
    
    def get_Kpanama(self):
        """
        Returns:
            Kpanama (normalized XX.T)
        """
        return self.Kpanama
    
    def get_Ktot(self):
        """
        Returns:
            Ktot (normalized Kpanama+Kpop)
        """
        return self.Ktot
    
    def get_varianceComps(self):
        """
        Returns:
            vector of variance components of the PANAMA, Kpop and noise contributions
        """
        return self.varianceComps




import unittest

class TestModules(unittest.TestCase):
    def test_variance_decomposition_import(self):
        try:
            import limix.modules.varianceDecomposition as vd
        except ImportError:
            self.fail()

if __name__ == '__main__':
    unittest.main()



from .covar_base import Covariance
import pdb
import numpy as np
import scipy as sp
from hcache import Cached, cached

class ACombinatorCov(Covariance):
    """
    Abstract class for combinator of covariances
    """
    def __init__(self):
        Covariance.__init__(self)
        self.covars = []

    #####################
    # Covars handling
    #####################
    def addCovariance(self,covar):
        if self.dim is None:
            self.dim = covar.dim
        else:
            assert covar.dim==self.dim, 'Dimension mismatch.'
        self.covars.append(covar)
        covar.register(self.clear_all)
        self._calcNumberParams()

    def getCovariance(self,i):
        return self.covars[i]

    #####################
    # Params handling
    #####################
    def setParams(self,params):
        istart = 0
        cs = [c for c in self.covars if c.getNumberParams() > 0]
        for c in cs:
            n = c.getNumberParams()
            istop = istart + n
            c.setParams(params[istart:istop])
            istart = istop

    def getParams(self):
        istart = 0
        params = sp.zeros(self.getNumberParams())
        cs = [c for c in self.covars if c.getNumberParams() > 0]
        for c in cs:
            istop = istart + c.getNumberParams()
            params[istart:istop] = c.getParams()
            istart = istop
        return params

    def getNumberParams(self):
        return np.sum([c.getNumberParams() for c in self.covars])

    ####################
    # Predictions
    ####################
    @property
    def use_to_predict(self):
        r = False
        for i in range(len(self.covars)):
            r = r or self.getCovariance(i).use_to_predict
        return r

    @use_to_predict.setter
    def use_to_predict(self,value):
        raise NotImplementedError("This method is only implemented for single"
                                  " covariance terms.")

    ####################
    # Interpretable Params and Ste
    ####################
    def getInterParams(self):
        istart = 0
        interParams = sp.zeros(self.getNumberParams())
        for i in range(len(self.covars)):
            istop = istart + self.getCovariance(i).getNumberParams()
            params[istart:istop] = self.getCovariance(i).getInterParams()
            istart = istop
        return params

    def setFIinv(self, value):
        self._FIinv = value
        istart = 0
        for i in range(len(self.covars)):
            istop = istart + self.getCovariance(i).getNumberParams()
            self.getCovariance(i).setFIinv(value[istart:istop][:,istart:istop])
            istart = istop





import sys
from limix.utils.preprocess import remove_dependent_cols
from limix.utils.util_functions import smartDumpDictHdf5

# core
from limix.core.gp import GP2KronSum
from limix.core.gp import GP2KronSumLR
from limix.core.gp import GP3KronSumLR
from limix.core.covar import FreeFormCov

import h5py
import pdb
import scipy as sp
import scipy.linalg as la
import scipy.stats as st
import time as TIME
import copy
import warnings
import os


class MTSet():
    """
    This class eases implementation of efficient set test algorithms.
    mtSet can account for relatedness and can be used for single- and multi-trait analysis
    """

    def __init__(self, Y=None, R=None, S_R=None, U_R=None, traitID=None, F=None, rank = 1):
        """
        Args:
            Y:          [N, P] phenotype matrix
            F:          [N, K] matrix of fixed effect design.
                        K is the number of per-trait covariates.
            R:          [N, N] genetic relatedness matrix between individuals.
                        In alternative to R, S_R and U_R can be provided.
                        If not specified a model without relatedness component is considered.
            S_R:        N vector of eigenvalues of R
            U_R:        [N, N] eigenvector matrix of R
            traiID:     P vector of the IDs of the phenotypes to analyze (optional)
            rank:       rank of the trait covariance matrix of the variance component to be tested (default is 1)
        """
        # data
        noneNone = S_R is not None and U_R is not None
        self.bgRE = R is not None or noneNone
        # fixed effect
        msg = 'The current implementation of the full rank mtSet'
        msg+= ' does not support covariates.'
        msg+= ' We reccommend to regress out covariates and'
        msg+= ' subsequently quantile normalize the phenotypes'
        msg+= ' to a normal distribution prior to use mtSet.'
        msg+= ' This can be done within the LIMIX framework using'
        msg+= ' the methods limix.utils.preprocess.regressOut and'
        msg+= ' limix.utils.preprocess.gaussianize'
        assert not (F is not None and self.bgRE), msg
        if F is not None:
            F = remove_dependent_cols(F)
            A = sp.eye(Y.shape[1])
        else:
            A = None
        #traitID
        if traitID is None:
            traitID = sp.array(['trait %d' % p for p in range(Y.shape[1])])
        self.setTraitID(traitID)
        #init covariance matrices and gp
        Cg = FreeFormCov(Y.shape[1])
        Cn = FreeFormCov(Y.shape[1])
        G  = 1. * (sp.rand(Y.shape[0],1)<0.2)
        if self.bgRE:
            self._gp = GP3KronSumLR(Y=Y, Cg=Cg, Cn=Cn, R=R, S_R=S_R, U_R=U_R, G=G, rank = rank)
        else:
            self._gp = GP2KronSumLR(Y=Y, Cn=Cn, G=G, F=F, A=A)
        # null model params
        self.null = None
        # calls itself for column-by-column trait analysis
        self.stSet = None
        self.nullST = None
        self.infoOpt   = None
        self.infoOptST = None
        pass

    ##################################################
    # Properties
    ##################################################
    @property
    def N(self):    return self._gp.covar.dim_r

    @property
    def P(self):    return self._gp.covar.dim_c

    @property
    def Cr(self):   return self._gp.covar.Cr

    @property
    def Cg(self):   return self._gp.covar.Cg

    @property
    def Cn(self):   return self._gp.covar.Cn

    @property
    def Y(self):    return self._gp.mean.Y

    @property
    def F(self):
        try:
            return self._gp.mean.F[0]
        except:
            return None

    @property
    def A(self):
        try:
            return self._gp.mean.A[0]
        except:
            return None

    @property
    def S_R(self):
        if self.bgRE:   RV = self._gp.covar.Sr()
        else:           RV = None
        return RV

    @property
    def U_R(self):
        if self.bgRE:   RV = self._gp.covar.Lr().T
        else:           RV = None
        return RV

    #########################################
    # Setters
    #########################################
    @Y.setter
    def Y(self, value):
        assert value.shape[0]==self.N, 'Dimension mismatch'
        assert value.shape[1]==self.P, 'Dimension mismatch'
        self._gp.mean.Y = Y

    def setTraitID(self,traitID):
        self.traitID = traitID

    ################################################
    # Fitting null model
    ###############################################
    def fitNull(self, verbose=False, cache=False, out_dir='./cache', fname=None, rewrite=False, seed=None, n_times=10, factr=1e3, init_method=None):
        """
        Fit null model
        """
        if seed is not None:    sp.random.seed(seed)

        read_from_file = False
        if cache:
            assert fname is not None, 'MultiTraitSetTest:: specify fname'
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            out_file = os.path.join(out_dir,fname)
            read_from_file = os.path.exists(out_file) and not rewrite

        RV = {}
        if read_from_file:
            f = h5py.File(out_file,'r')
            for key in list(f.keys()):
                RV[key] = f[key][:]
            f.close()
            self.setNull(RV)
        else:
            start = TIME.time()
            if self.bgRE:
                self._gpNull = GP2KronSum(Y=self.Y, F=None, A=None, Cg=self.Cg, Cn=self.Cn, R=None, S_R=self.S_R, U_R=self.U_R)
            else:
                self._gpNull = GP2KronSumLR(self.Y, self.Cn, G=sp.ones((self.N,1)), F=self.F, A=self.A)
                # freezes Cg to 0
                n_params = self._gpNull.covar.Cr.getNumberParams()
                self._gpNull.covar.Cr.setParams(1e-9 * sp.ones(n_params))
                self._gpNull.covar.act_Cr = False
            for i in range(n_times):
                params0 = self._initParams(init_method=init_method)
                self._gpNull.setParams(params0)
                conv, info = self._gpNull.optimize(verbose=verbose, factr=factr)
                if conv: break
            if not conv:    warnings.warn("not converged")
            LMLgrad = (self._gpNull.LML_grad()['covar']**2).mean()
            LML = self._gpNull.LML()
            if self._gpNull.mean.n_terms==1:
                RV['B'] = self._gpNull.mean.B[0]
            elif self._gpNull.mean.n_terms>1:
                warning.warn('generalize to more than 1 fixed effect term')
            if self.bgRE:
                RV['params0_g'] = self.Cg.getParams()
            else:
                RV['params0_g'] = sp.zeros_like(self.Cn.getParams())
            RV['params0_n'] = self.Cn.getParams()
            if self.bgRE:
                RV['Cg'] = self.Cg.K()
            else:
                RV['Cg'] = sp.zeros_like(self.Cn.K())
            RV['Cn'] = self.Cn.K()
            RV['conv'] = sp.array([conv])
            RV['time'] = sp.array([TIME.time()-start])
            RV['NLL0'] = sp.array([LML])
            RV['LMLgrad'] = sp.array([LMLgrad])
            RV['nit'] = sp.array([info['nit']])
            RV['funcalls'] = sp.array([info['funcalls']])
            self.null = RV
            if cache:
                f = h5py.File(out_file,'w')
                smartDumpDictHdf5(RV,f)
                f.close()
        return RV

    def getNull(self):
        """ get null model info """
        return self.null

    def setNull(self,null):
        """ set null model info """
        self.null = null

    ###########################################
    # Fitting alternative model
    ###########################################

    def optimize(self, G, params0=None, n_times=10, verbose=False, vmax=5, perturb=1e-3, factr=1e7):
        """
        Optimize the model considering G
        """
        # set params0 from null if params0 is None
        if params0 is None:
            if self.null is None:
                if verbose:     print(".. fitting null model")
                self.fitNull()
            if self.bgRE:
                params0 = sp.concatenate([self.null['params0_g'], self.null['params0_n']])
            else:
                params0 = self.null['params0_n']
            params_was_None = True
        else:
            params_was_None = False
        G *= sp.sqrt(self.N/(G**2).sum())
        self._gp.covar.G = G
        start = TIME.time()
        for i in range(n_times):
            if params_was_None:
                n_params = self.Cr.getNumberParams()
                _params0 = {'covar': sp.concatenate([1e-3*sp.randn(n_params), params0])}
            else:
                _params0 = {'covar': params0}
            self._gp.setParams(_params0)
            conv, info = self._gp.optimize(factr=factr, verbose=verbose)
            conv *= self.Cr.K().diagonal().max()<vmax
            conv *= self.getLMLgrad() < 0.1
            if conv or not params_was_None: break
        self.infoOpt = info
        if not conv:
            warnings.warn("not converged")
        # return value
        RV = {}
        if self.P>1:
            RV['Cr']  = self.Cr.K()
            if self.bgRE: RV['Cg']  = self.Cg.K()
            RV['Cn']  = self.Cn.K()
        RV['time']  = sp.array([TIME.time()-start])
        RV['params0'] = _params0
        RV['nit'] = sp.array([info['nit']])
        RV['funcalls'] = sp.array([info['funcalls']])
        RV['var']    = self.getVariances()
        RV['conv']  = sp.array([conv])
        RV['NLLAlt']  = sp.array([self.getNLLAlt()])
        RV['LLR']    = sp.array([self.getLLR()])
        RV['LMLgrad'] = sp.array([self.getLMLgrad()])
        return RV

    def getInfoOpt(self):
        """ get information for the optimization """
        return self.infoOpt

    def getVariances(self):
        """
        get variances
        """
        var = []
        var.append(self.Cr.K().diagonal())
        if self.bgRE:
            var.append(self.Cg.K().diagonal())
        var.append(self.Cn.K().diagonal())
        var = sp.array(var)
        return var

    def getNLLAlt(self):
        """
        get negative log likelihood of the alternative
        """
        return self._gp.LML()

    def getLLR(self):
        """
        get log likelihood ratio
        """
        assert self.null is not None, 'null model needs to be fitted!'
        return self.null['NLL0'][0] - self.getNLLAlt()

    def getLMLgrad(self):
        """
        get norm LML gradient
        """
        return (self._gp.LML_grad()['covar']**2).mean()

    def fitNullTraitByTrait(self, verbose=False, cache=False, out_dir='./cache', fname=None, rewrite=False):
        """
        Fit null model trait by trait
        """
        read_from_file = False
        if cache:
            assert fname is not None, 'MultiTraitSetTest:: specify fname'
            if not os.path.exists(out_dir): os.makedirs(out_dir)
            out_file = os.path.join(out_dir,fname)
            read_from_file = os.path.exists(out_file) and not rewrite

        RV = {}
        if read_from_file:
            f = h5py.File(out_file,'r')
            for p in range(self.P):
                trait_id = self.traitID[p]
                g = f[trait_id]
                RV[trait_id] = {}
                for key in list(g.keys()):
                    RV[trait_id][key] = g[key][:]
            f.close()
            self.nullST=RV
        else:
            """ create stSet and fit null column by column returns all info """
            if self.stSet is None:
                y = sp.zeros((self.N,1))
                self.stSet = MTSet(Y=y, S_R=self.S_R, U_R=self.U_R, F=self.F)
            RV = {}
            for p in range(self.P):
                trait_id = self.traitID[p]
                self.stSet.Y = self.Y[:,p:p+1]
                RV[trait_id] = self.stSet.fitNull()
            self.nullST = RV
            if cache:
                f = h5py.File(out_file,'w')
                smartDumpDictHdf5(RV,f)
                f.close()
        return RV

    def optimizeTraitByTrait(self, G, verbose=False, n_times=10, factr=1e3):
        """ Optimize trait by trait """
        assert self.nullST is not None, 'fit null model beforehand'
        RV = {}
        self.infoOptST = {}
        for p in range(self.P):
            trait_id = self.traitID[p]
            self.stSet.Y = self.Y[:, p:p+1]
            self.stSet.setNull(self.nullST[trait_id])
            RV[trait_id] = self.stSet.optimize(G, n_times=n_times, factr=factr, verbose=verbose)
            self.infoOptST[trait_id] = self.stSet.getInfoOpt()
        return RV

    def getInfoOptST(self):
        """ get information for the optimization """
        return self.infoOptST

    def _initParams(self, init_method=None):
        """ this function initializes the paramenter and Ifilter """
        if self.bgRE:
            if init_method=='random':
                params0 = {'covar': sp.randn(self._gpNull.covar.getNumberParams())}
            else:
                if self.P==1:
                    params0 = {'covar':sp.sqrt(0.5) * sp.ones(2)}
                else:
                    cov = 0.5*sp.cov(self.Y.T) + 1e-4*sp.eye(self.P)
                    chol = la.cholesky(cov, lower=True)
                    params = chol[sp.tril_indices(self.P)]
                    params0 = {'covar': sp.concatenate([params, params])}
        else:
            if self.P==1:
                params_cn = sp.array([1.])
            else:
                cov = sp.cov(self.Y.T) + 1e-4*sp.eye(self.P)
                chol = la.cholesky(cov, lower=True)
                params_cn = chol[sp.tril_indices(self.P)]
            params0 = {'covar': params_cn}
        return params0

if __name__=='__main__':
    from limix.utils.preprocess import covar_rescale

    sp.random.seed(0)

    #generate phenotype
    n = 1000
    p = 4
    f = 10
    Y = sp.randn(n, p)
    X = sp.randn(n, f)
    G = sp.randn(n, f)
    R = sp.dot(X, X.T)
    R = covar_rescale(R)

    mts = mtset(Y, R=R)
    nullMTInfo = mts.fitNull(cache=False)
    mts.optimize(G)



"""LMM testing code"""
import unittest
import scipy as sp
import numpy as np
import sys
from limix.core.covar import Cov2KronSum 
from limix.core.covar import Cov3KronSumLR
from limix.core.covar import FreeFormCov
from limix.utils.check_grad import mcheck_grad
from limix.utils.preprocess import covar_rescale
from limix.core.type.exception import TooExpensiveOperationError

class TestCov3KronSumLR(unittest.TestCase):
    def setUp(self):
        sp.random.seed(1)
        # define row caoriance
        dim_r = 4
        rank_r = 2
        G = sp.rand(dim_r, rank_r)
        X = sp.rand(dim_r, dim_r)
        R = covar_rescale(sp.dot(X,X.T))
        # define col covariances
        dim_c = 2
        Cg = FreeFormCov(dim_c)
        Cn = FreeFormCov(dim_c)
        self.C = Cov3KronSumLR(Cn = Cn, Cg = Cg, R = R, G = G, rank = 1)
        self.name = 'cov3kronSumLR'
        self.C.setRandomParams()

    def test_solve(self):
        x = sp.randn(self.C.dim, 1)
        v1 = self.C.solve(x)
        v2 = Cov2KronSum.solve(self.C, x)
        np.testing.assert_almost_equal(v1, v2, decimal = 5)

    def test_logdet_grad(self):
        def func(x, i):
            self.C.setParams(x)
            return self.C.logdet()

        def grad(x, i):
            self.C.setParams(x)
            return self.C.logdet_grad_i(i)

        x0 = self.C.getParams()
        err = mcheck_grad(func, grad, x0)
        np.testing.assert_almost_equal(err, 0., decimal = 5)

    def test_too_expensive_exceptions(self):
        dim_r = 100
        rank_r = 2
        G = sp.rand(dim_r, rank_r)
        X = sp.rand(dim_r, dim_r)
        R = covar_rescale(sp.dot(X,X.T))

        dim_c = 5001
        Cg = FreeFormCov(dim_c)
        Cn = FreeFormCov(dim_c)
        C = Cov3KronSumLR(Cn = Cn, Cg = Cg, R = R, G = G, rank = 1)

        with self.assertRaises(TooExpensiveOperationError):
            C.L()
        with self.assertRaises(TooExpensiveOperationError):
            C.K()
        with self.assertRaises(TooExpensiveOperationError):
            C.K_grad_i(0)

    def test_param_activation(self):
        self.C.act_Cr = False
        self.C.act_Cg = False
        self.C.act_Cn = False
        self.assertEqual(len(self.C.getParams()), 0)

        with self.assertRaises(ValueError):
            self.C.K_grad_i(0)

    def test_H_chol(self):
        C = self.C

        R  = sp.dot(C.W().T, C.DW())
        R += sp.eye(R.shape[0])
        chol1 = np.linalg.cholesky(R)
        chol2 = C.H_chol()

        np.testing.assert_array_almost_equal(chol1, chol2)

    def test_inv(self):
        C = self.C

        dL = C.d()[:,sp.newaxis] * C.L()
        WdL = sp.dot(C.DW().T, C.L())
        HiWdL = sp.dot(C.H_inv(), WdL)
        inv1 = sp.dot(C.L().T, dL) - sp.dot(WdL.T, HiWdL)
        inv2 = C.inv()

        np.testing.assert_array_almost_equal(inv1, inv2)

    def test_logdet(self):
        d1 = 2*sp.log(sp.diag(self.C.chol())).sum()
        d2 = self.C.logdet()
        np.testing.assert_almost_equal(d1, d2)

if __name__ == '__main__':
    unittest.main()



"""GP testing code"""
import unittest
import scipy as SP
import numpy as np
import limix.deprecated as dlimix
import scipy.linalg as linalg


def PCA(Y, components):
    """run PCA, retrieving the first (components) principle components
    return [s0, eig, w0]
    s0: factors
    w0: weights
    """
    sv = linalg.svd(Y, full_matrices=0);
    [s0, w0] = [sv[0][:, 0:components], SP.dot(SP.diag(sv[1]), sv[2]).T[:, 0:components]]
    v = s0.std(axis=0)
    s0 /= v;
    w0 *= v;
    return [s0, w0]



class CGPLVM_test(unittest.TestCase):
    """oGPLVM test class"""

    def simulate(self):
        """simulate a dataset. Note this is seed-dependent"""

        N = self.settings['N']
        K = self.settings['K']
        D = self.settings['D']

        SP.random.seed(1)
        S = SP.random.randn(N,K)
        W = SP.random.randn(D,K)

        Y = SP.dot(W,S.T).T
        Y+= 0.1*SP.random.randn(N,D)

        X0 = SP.random.randn(N,K)
        X0 = PCA(Y,K)[0]
        RV = {'X0': X0,'Y':Y,'S':S,'W':W}
        return RV

    def setUp(self):
        SP.random.seed(1)

        #1. simulate
        self.settings = {'K':5,'N':100,'D':80}
        self.simulation = self.simulate()

        N = self.settings['N']
        K = self.settings['K']
        D = self.settings['D']

        #2. setup GP
        covar  = dlimix.CCovLinearISO(K)
        ll  = dlimix.CLikNormalIso()
        #create hyperparm
        covar_params = SP.array([1.0])
        lik_params = SP.array([1.0])
        hyperparams = dlimix.CGPHyperParams()
        hyperparams['covar'] = covar_params
        hyperparams['lik'] = lik_params
        hyperparams['X']   = self.simulation['X0']
        #cretae GP
        self.gp=dlimix.CGPbase(covar,ll)
        #set data
        self.gp.setY(self.simulation['Y'])
        self.gp.setX(self.simulation['X0'])
        self.gp.setParams(hyperparams)
        pass

    @unittest.skip("someone has to fix it")
    def test_fit(self):
        #create optimization object
        self.gpopt = dlimix.CGPopt(self.gp)
        #run
        RV = self.gpopt.opt()
        RV = self.gpopt.opt()

        m = (SP.absolute(self.gp.LMLgrad()['X']).max() +
             SP.absolute(self.gp.LMLgrad()['covar']).max() +
             SP.absolute(self.gp.LMLgrad()['lik']).max())

        np.testing.assert_almost_equal(m, 0., decimal=1)


class CGPLVM_test_constK(unittest.TestCase):
    """adapted version of GPLVM test, including a fixed CF covaraince"""

    def simulate(self):
        """simulate a dataset. Note this is seed-dependent"""

        N = self.settings['N']
        K = self.settings['K']
        D = self.settings['D']

        SP.random.seed(1)
        S = SP.random.randn(N,K)
        W = SP.random.randn(D,K)

        Y = SP.dot(W,S.T).T
        Y+= 0.1*SP.random.randn(N,D)

        X0 = SP.random.randn(N,K)
        X0 = PCA(Y,K)[0]
        RV = {'X0': X0,'Y':Y,'S':S,'W':W}
        return RV

    def setUp(self):
        SP.random.seed(1)

        #1. simulate
        self.settings = {'K':5,'N':100,'D':80}
        self.simulation = self.simulate()

        N = self.settings['N']
        K = self.settings['K']
        D = self.settings['D']

        #2. setup GP
        K0 = SP.dot(self.simulation['S'],self.simulation['S'].T)
        K0[:] = 0

        covar1 = dlimix.CFixedCF(K0)
        covar2 = dlimix.CCovLinearISO(K)
        covar  = dlimix.CSumCF()
        covar.addCovariance(covar1)
        covar.addCovariance(covar2)

        ll  = dlimix.CLikNormalIso()
        #create hyperparm
        covar_params = SP.array([0.0,1.0])
        lik_params = SP.array([0.1])
        hyperparams = dlimix.CGPHyperParams()
        hyperparams['covar'] = covar_params
        hyperparams['lik'] = lik_params
        hyperparams['X']   = self.simulation['X0']
        #cretae GP
        self.gp=dlimix.CGPbase(covar,ll)
        #set data
        self.gp.setY(self.simulation['Y'])
        self.gp.setX(self.simulation['X0'])
        self.gp.setParams(hyperparams)
        pass

    @unittest.skip("someone has to fix it")
    def test_fit(self):
        #create optimization object
        self.gpopt = dlimix.CGPopt(self.gp)
        #run
        RV = self.gpopt.opt()
        RV = self.gpopt.opt()

        m = (SP.absolute(self.gp.LMLgrad()['X']).max() +
             SP.absolute(self.gp.LMLgrad()['covar']).max() +
             SP.absolute(self.gp.LMLgrad()['lik']).max())

        np.testing.assert_almost_equal(m, 0., decimal=1)


if __name__ == '__main__':
    unittest.main()



import scipy as SP
import pdb
import scipy.stats as STATS

class Chi2mixture(object):

    def __init__(self, scale_min=0.1,scale_max=5.0,dof_min=0.1,dof_max=5.0,n_intervals=100,qmax=0.1,tol=0):
        """
        class for a mixture of chi2-distributed random variables of the form

        (1-mixture)*chi2(0) + (mixture)*scale*chi2(dof),

        where
        scale   is the scaling parameter for the scales chi-square distribution
        dof     are the degrees of freedom of the second component
        mixture is the probability of beeing in the first component

        input:
        scale_min   minimum value used for fitting the scale parameter
        scale_max   maximum value used for fitting the scale parameter
        dofmin      minimum value used for fitting the dof parameter
        dofmax      maximum value used for fitting the dof parameter
        qmax        only the top qmax quantile is used for the fit
        n_interval  number of intervals when performing gridsearch
        tol        tolerance of being zero
        """
        
        self.scale_min   = scale_min
        self.scale_max   = scale_max
        self.dof_min     = dof_min
        self.dof_max     = dof_max
        self.qmax        = qmax
        self.n_intervals = n_intervals
        self.tol         = tol
    
    def estimate_chi2mixture(self, lrt):
        """
        estimates the parameters of a mixture of a chi-squared random variable of degree
        0 and a scaled chi-squared random variable of degree d
        
        (1-mixture)*chi2(0) + (mixture)*scale*chi2(dof),

        where
        scale   is the scaling parameter for the scales chi-square distribution
        dof     are the degrees of freedom of the second component
        mixture is the probability of beeing in the first component

        input:
        lrt         [Ntests] vector of test statistics
        """

        """
        step 1: estimate the probability of being in component one
        """
        self.mixture = 1-(lrt<=self.tol).mean()
        n_false      = SP.sum(lrt>self.tol)

        """
        step 2: only use the largest qmax fraction of test statistics to estimate the
                remaining parameters
        """
        n_fitting   = SP.ceil(self.qmax * n_false)
        lrt_sorted  = -SP.sort(-lrt)[:n_fitting]
        q           = SP.linspace(0, 1,n_false)[1:n_fitting+1]
        log_q       = SP.log10(q)
        
        """
        step 3: fitting scale and dof by minimizing the squared error of the log10 p-values
                with their theorietical values [uniform distribution]
        """
        MSE_opt = SP.inf
        MSE     = SP.zeros((self.n_intervals,self.n_intervals))

        
        for i,scale in enumerate(SP.linspace(self.scale_min,self.scale_max,self.n_intervals)):
            for j,dof in enumerate(SP.linspace(self.dof_min,self.dof_max,self.n_intervals)):
                p     = STATS.chi2.sf(lrt_sorted/scale,dof)
                log_p = SP.log10(p)
                MSE[i,j]   = SP.mean((log_q - log_p)**2)
                if MSE[i,j] < MSE_opt:
                    MSE_opt    = MSE[i,j]
                    self.scale = scale
                    self.dof   = dof

        
    
    def sf(self,lrt):
        """
        computes the survival function of a mixture of a chi-squared random variable of degree
        0 and a scaled chi-squared random variable of degree d
        """
        _lrt = SP.copy(lrt)
        _lrt[lrt<self.tol] = 0
        pv = self.mixture*STATS.chi2.sf(_lrt/self.scale,self.dof)
        return pv
        
if __name__ == "__main__":
    scale = 0.3
    dof = 2
    mixture = 0.2
    n_test = 100000
    n_train= 100000

    lrt_train = SP.zeros((n_train))
    lrt_test  = SP.zeros((n_test))
    
    for i in range(dof):
        x = SP.random.randn(n_train)
        lrt_train += scale * (x*x)
        x = SP.random.randn(n_test)
        lrt_test  += scale * (x*x)

    idx_chi2_0      = SP.random.permutation(n_train)[:(1-mixture)*n_train]
    lrt_train[idx_chi2_0] = 1e-10*SP.random.randn((1-mixture)*n_train)
    
    idx_chi2_0      = SP.random.permutation(n_test)[:(1-mixture)*n_test]
    lrt_test[idx_chi2_0] = 1e-10*SP.random.randn((1-mixture)*n_test)


    chi2mix = Chi2mixture(scale_min=0.1, scale_max=5.0, dof_min=0.1, dof_max=5.0, qmax=0.1,tol=4e-3)
    chi2mix.estimate_chi2mixture(lrt_train)
    chi2mix.sf(lrt_test)

    print(('true scale = %.2f'%scale))
    print(('true dof   = %.2f'%dof))
    print(('true mixt  = %.2f'%mixture))
    print(('est scale  = %.2f'%chi2mix.scale))
    print(('est dof    = %.2f'%chi2mix.dof))
    print(('est mixt   = %.2f'%chi2mix.mixture))
    
    import matplotlib.pylab as PLT
    fig = PLT.figure()
    fig.add_subplot(121)
    PLT.hist(chi2mix.sf(lrt_train),normed=True)
    fig.add_subplot(122)
    PLT.hist(chi2mix.sf(lrt_test),normed=True)

    pval_test = chi2mix.sf(lrt_test)
    print('print Type 1 error estimate')
    alpha = 1e-5
    print('... alpha=1e-5')
    print(('...... type 1 error: %.2e'%(pval_test<alpha).mean()))
    alpha = 1e-4
    print('... alpha=1e-4')
    print(('...... type 1 error: %.2e'%(pval_test<alpha).mean()))
    alpha = 1e-3
    print('... alpha=1e-3')
    print(('...... type 1 error: %.2e'%(pval_test<alpha).mean()))



import sys
from limix.core.type.observed import Observed
from hcache import Cached, cached
from limix.utils.preprocess import regressOut
from limix.core.utils import assert_finite_array
import scipy as sp
import numpy as np

import scipy.linalg as LA
import copy

class MeanBase(Cached, Observed):
    """
    Basic mean function for gp regression
    Notation:
        N = number of individuals
        No = number of out-of-sample individuals for predictions
        K = number of fixed effect covariates
    """

    def __init__(self, y, W=None, Wstar=None):
        """
        Args:
            y:        phenotype vector [N, 1]
            W:        fixed effect design [N, K]
            Wstar:    out-of-sample fixed effect design for predictions [No, K]
        """
        Cached.__init__(self)
        self.y = y
        self.W = W
        self.Wstar = Wstar
        self.setFIinv(None)
        self._set_relay(None)

    def _set_relay(self, relay):
        self._relay = relay

    #########################################
    # Properties
    #########################################
    @property
    def W(self):
        return self._W

    @property
    def n_covs(self):
        return self.W.shape[1]

    @property
    def y(self):
        return self._y

    @property
    def b(self):
        return self._relay.b()

    @property
    def b_ste(self):
        if self.getFIinv() is None:
            R = None
        else:
            R = sp.sqrt(self.getFIinv().diagonal())[:,sp.newaxis]
        return R

    @property
    def n_data_points(self):
        return self._N

    # for multitrait interpretation
    @property
    def Y(self):
        return self.y

    @property
    def B(self):
        return self.b

    @property
    def B_ste(self):
        return self.b_ste

    @property
    def Wstar(self):
        return self._Wstar

    @property
    def use_to_predict(self):
        return self._use_to_predict

    #########################################
    # Setters
    #########################################
    #@Y.setter
    #def Y(self,value):
    #    """ set phenotype """
    #    self._N = value.shape[0]
    #    self._P = value.shape[1]
    #    self._Y = value
    #    self._notify()
    #    self.clear_cache('Yres')

    @W.setter
    def W(self,value):
        """ set fixed effect design """
        if value is None:   value = sp.zeros((self._N, 0))
        assert value.shape[0]==self._N, 'Dimension mismatch'
        self._K = value.shape[1]
        self._W = value
        self._notify()
        self.clear_cache('predict_in_sample','Yres')

    @Wstar.setter
    def Wstar(self,value):
        """ set fixed effect design for predictions """
        if value is None:
            self._use_to_predict = False
        else:
            assert value.shape[1]==self._K, 'Dimension mismatch'
            self._use_to_predict = True
        self._Wstar = value
        self.clear_cache('predict')

    @y.setter
    def y(self,value):
        assert_finite_array(value)
        assert value.shape[1] == 1, 'MeanBase: phenotype has to be a one column vector'
        self._N = value.shape[0]
        self._P = 1
        self._y = value
        # notify
        self._notify()
        self.clear_cache('Yres')

    @use_to_predict.setter
    def use_to_predict(self, value):
        assert not (self.Wstar is None and value is True), 'set Wstar!'
        self._use_to_predict = value

    #########################################
    # Predictions
    #########################################
    @cached
    def predict(self):
        r = _predict_fun(self.Wstar)
        return r

    @cached
    def predict_in_sample(self):
        r = _predict_fun(self.W)
        return r

    def _predict_fun(self,M):
        return sp.dot(M,self.b)

    def Yres(self):
        return self.yres

    @cached('Yres')
    def yres(self):
        return self.y - self.predict_in_sample()

    #######################################
    # Standard errors
    ########################################
    def setFIinv(self, value):
        self._FIinv = value

    def getFIinv(self):
        return self._FIinv

    ###########################################
    # Gradient TODO
    ###########################################
    #def getGradient(self,j):
    #    """ get gradient for fixed effect i """
    #    return rv

    #########################################
    # Params manipulation TODO
    #########################################
    #def getParams(self):
    #    """ get params """
    #    return rv

    #def setParams(self,params):
    #    """ set params """
    #    start = 0
    #    for i in range(self.n_terms):
    #        n_effects = self.B[i].size
    #        self.B[i] = np.reshape(params[start:start+n_effects],self.B[i].shape, order='F')
    #        start += n_effects



__author__ = 'damienarnol1'

import os
import shutil
import glob
import numpy as np
import scipy.spatial as SS



def make_dir(dir_name, overwrite=False):
    while dir_name[-1] == '/':
        dir_name = dir_name[0:-1]

    if overwrite:
        if os.path.exists(dir_name):
            shutil.rmtree(dir_name)

    else:
        count = 2
        dir_name_tmp = dir_name
        while os.path.exists(dir_name_tmp):
            dir_name_tmp = dir_name+str(count)
            count += 1
        dir_name = dir_name_tmp

    os.makedirs(dir_name)
    return dir_name+'/'


def compute_all_pairwise_distances(positions_directory, output_directory):
    all_files = sorted(glob.glob(positions_directory+'/*'))

    counter = 0

    for file_name in all_files:
        positions = np.loadtxt(file_name, delimiter=',')
        distances = SS.distance.pdist(positions, 'euclidean')

        image_name = file_name.split('/')[-1]
        output_file_name = output_directory+'/'+image_name
        np.savetxt(output_file_name, distances, delimiter=' ')








import scipy as sp
import numpy as np
import scipy.linalg as la
from hcache import cached
from limix.core.utils import assert_make_float_array
from limix.core.utils import assert_finite_array
from .covar_base import Covariance
import pdb

class FixedCov(Covariance):
    """
    Fixed-form covariance matrix.
    A fixed-form covariance matrix has only 1 parameter
    """
    def __init__(self, K0, Kcross0=None):
        """
        Args:
            K0:         semi-definite positive matrix that defines the fixed-form covariance
            Kcross0:    cross covariance between training and test samples
                        (used only for out-of-sample predictions)
        """
        Covariance.__init__(self)
        self._scale_act = True
        self.K0 = assert_make_float_array(K0, "K0")
        assert_finite_array(self.K0)

        if Kcross0 is not None:
            Kcross0 = assert_make_float_array(Kcross0, "Kcross0")
            assert_finite_array(Kcross0)

        self.Kcross0 = Kcross0
        self.scale = 1

    #####################
    # Properties
    #####################
    @property
    def scale(self):
        return self._scale 

    @property
    def scale_ste(self):
        if self.getFIinv() is None:
            R = None
        else:
            R = sp.sqrt(self.getFIinv()[0,0])
        return R

    @property
    def K0(self):
        return self._K0

    @property
    @cached('K0')
    def X0(self):
        S, U = la.eigh(self.K0)
        I = (S>1e-9)
        return U[:,I]*S[I]**(0.5) 

    @property
    @cached(['covar_base', 'K0'])
    def X(self):
        return sp.sqrt(self.scale) * self.X0

    @property
    def Kcross0(self):
        return self._Kcross0

    #####################
    # Setters
    #####################
    @scale.setter
    def scale(self,value):
        assert value >= 0, 'Scale must be >= 0.'
        self._scale = value 
        self.clear_all()

    @K0.setter
    def K0(self,value):
        self._K0 = value
        self.initialize(value.shape[0])
        self.clear_cache('K0')
        self._notify()

    @Kcross0.setter
    def Kcross0(self,value):
        if value is None:
            self._use_to_predict = False
        else:
            assert value.shape[1] == self.dim, 'Dimension mismatch.'
            self._use_to_predict = True
        self._Kcross0 = value
        self.clear_cache('Kcross0')
        self._notify()

    @Covariance.use_to_predict.setter
    def use_to_predict(self,value):
        assert self.Kcross0 is not None, 'Kcross0 has to be set before.'
        self._use_to_predict = value

    #####################
    # Activation handling
    #####################
    @property
    def act_scale(self):
        return self._scale_act

    @act_scale.setter
    def act_scale(self, act):
        self._scale_act = bool(act)
        self._notify()

    #####################
    # Params handling
    #####################
    def setParams(self, params):
        if int(self._scale_act) != len(params):
            raise ValueError("The number of parameters passed to setParams "
                             "differs from the number of active parameters.")
        if self._scale_act:
            self.scale = sp.exp(params[0])
            self.clear_all()

    def _calcNumberParams(self):
        self.n_params = 1

    def getParams(self):
        if self._scale_act:
            return sp.array([sp.log(self.scale)])
        return np.array([])

    def getNumberParams(self):
        """
        return the number of hyperparameters
        """
        return int(self._scale_act)

    #####################
    # Cached
    #####################
    @cached(['covar_base', 'K0'])
    def K(self):
        return self.scale * self.K0

    @cached(['Kcross0', 'covar_base'])
    def Kcross(self):
        return self.scale * self.Kcross0

    @cached(['covar_base', 'K0'])
    def K_grad_i(self,i):
        if i >= int(self._scale_act):
            raise ValueError("Trying to retrieve the gradient over a "
                             "parameter that is inactive.")
        r = self.scale * self.K0
        return r

    @cached(['covar_base', 'K0'])
    def K_hess_i_j(self, i, j):
        if i >= int(self._scale_act) or j >= int(self._scale_act):
            raise ValueError("Trying to retrieve the hessian over a "
                             "parameter that is inactive.")
        return self.K()

    @cached(['covar_base', 'K0'])
    def Xgrad(self, i):
        return 0.5*self.X

    ####################
    # Interpretable Params
    ####################
    def getInterParams(self):
        if self._scale_act:
            return SP.array([self.scale])
        return np.array([])

    def K_grad_interParam_i(self,i):
        if i >= int(self._scale_act):
            raise ValueError("Trying to retrieve the gradient over a "
                             "parameter that is inactive.")

        return self.K0


if __name__=='__main__':

    import pdb
    C = FixedCov(sp.ones((2,2)))
    pdb.set_trace()




import sys

try:
    sys.path.insert(0,'./../../..')
    sys.path.insert(0,'./../../../../../mksum/mtSet')
    from mtSet.pycore.gp import gp2kronSum as gp2kronSumMtSet
    import mtSet.pycore.mean as MEAN
    mtSet_present = True
except:
    print("no mtSet found in path")
    mtSet_present = False

from limix.core.mean import mean
from limix.core.gp import gp2kronSum as gp2kronSum

from limix.core.covar import freeform 
import limix.core.optimize.optimize_bfgs as OPT


import scipy as SP
import scipy.linalg as LA 
import time as TIME
import copy


if __name__ == "__main__":

    # generate data
    h2 = 0.3
    N = 500; P = 3; S = 1000
    X = 1.*(SP.rand(N,S)<0.2)
    beta = SP.randn(S,P)
    Yg = SP.dot(X,beta); Yg*=SP.sqrt(h2/Yg.var(0).mean())
    Yn = SP.randn(N,P); Yn*=SP.sqrt((1-h2)/Yn.var(0).mean())
    Y  = Yg+Yn; Y-=Y.mean(0); Y/=Y.std(0)
    XX = SP.dot(X,X.T)
    XX/= XX.diagonal().mean()
    Xr = 1.*(SP.rand(N,10)<0.2)
    Xr*= SP.sqrt(N/(Xr**2).sum())

    # define mean term
    mu = mean(Y)

    # add first fixed effect
    F = 1.*(SP.rand(N,2)<0.2); A = SP.eye(P)

    # add second fixed effect
    F2 = 1.*(SP.rand(N,3)<0.2); A2 = SP.ones((1,P))
    
    mu.addFixedEffect(F=F,A=A)
    mu.addFixedEffect(F=F2,A=A2)

    if mtSet_present:
        mu1 = MEAN.mean(Y)
        mu1.addFixedEffect(F=F,A=A)
        mu1.addFixedEffect(F=F2,A=A2)
    

    # define covariance matrices
    Cg = freeform(P)
    Cn = freeform(P)
    Cg1 = freeform(P)
    Cn1 = freeform(P)
    
    ipdb.set_trace()

    if 1:
        # compare with mtSet implementation
        params = {}
        params['Cg']   = SP.randn(int(0.5*P*(P+1)))
        params['Cn']   = SP.randn(int(0.5*P*(P+1)))


        print("check gradient with gp2kronSum")
        gp = gp2kronSum(mu,Cg,Cn,XX)
        gp.setParams(params)

        if 0:
            gp.set_reml(False)

        print("test optimization")
        start = TIME.time()
        conv,info = OPT.opt_hyper(gp,params,factr=1e3)
        print(('Reml GP:', TIME.time()-start))
        
        if mtSet_present:
            params1 = copy.copy(params)
            params1['mean'] = SP.zeros(mu1.getParams().shape[0])
            gp1 = gp2kronSumMtSet(mu1,Cg1,Cn1,XX)
            gp1.setParams(params1)
            start = TIME.time()
            conv1,info = OPT.opt_hyper(gp1,params1,factr=1e3)
            print(('Old GP:', TIME.time()-start))

        print(conv)

        ipdb.set_trace()

    if 1:
        # no fixed 
        mu = mean(Y)
        params = {}
        params['Cg']   = SP.randn(int(0.5*P*(P+1)))
        params['Cn']   = SP.randn(int(0.5*P*(P+1)))

        ipdb.set_trace()
        gp = gp2kronSum(mu,Cg,Cn,XX)
        gp.setParams(params)
        gp.LML()
        gp.LMLgrad()




__author__ = 'damienarnol1'

from . import util_functions
from .run_individual_model import run_individual_model
import sys


def run_all_models(input_directory, replicate_index=None, overwrite=False):
    # create a results directory in which to write the files
    if replicate_index is not None:
        results_directory = input_directory + '/results_' + str(replicate_index) + '/'
        rd_sp = True
    else:
        results_directory = input_directory+'/results/'
        rd_sp =False
    results_directory = util_functions.make_dir(results_directory, overwrite)

    # file_names
    positions = input_directory+'/positions.txt'
    expressions = input_directory+'/expressions.txt'

    # full model
    run_individual_model('full', expressions, positions, results_directory,
                         permute_positions=False, random_start_point=rd_sp)

    # full model plus random
    run_individual_model('full', expressions, positions, results_directory,
                         permute_positions=True, random_start_point=rd_sp)

    # env model
    run_individual_model('env', expressions, positions, results_directory,
                         permute_positions=False, random_start_point=rd_sp)

    # env model plus random
    run_individual_model('env', expressions, positions, results_directory,
                         permute_positions=True, random_start_point=rd_sp)

    return



analysis_directory = sys.argv[1]
replicate_index = sys.argv[2]
run_all_models(analysis_directory, replicate_index, False)



#IGE (Indirect genetic effects) = SGE (Social genetic effects). IEE = SEE.

#the potential presence of NAs in input phenotype, covs, cages etc means that in this code we introduce two sets of animals (not discussed in the paper):
#focal animals, defined as having phenotype, covs (if covs are provided), cage and kinship
#all animals = focal animals + cage mates, where cage mates are defined as having cage and kinship and being in subset_IDs, and are referenced to using _cm in this code. Note that in the paper cage mate has a more precise meaning, namely "the other mice in the cage of the focal individual".

import sys
import warnings
import scipy as sp
import scipy.linalg as la
from limix.core.mean.mean_base import MeanBase as lin_mean
from limix.core.covar.dirIndirCov import DirIndirCov
from limix.core.covar.fixed import FixedCov 
from limix.core.covar.combinators import SumCov
from limix.utils.preprocess import covar_rescaling_factor
from limix.utils.preprocess import covar_rescale
from limix.core.gp.gp_base import GP
import pdb

class DirIndirVD():

    #many of inputs below are passed from SocialData object built using social_data.py
    #pheno: missing values should be encoded as -999 (needs to be numeric for HDF5). pheno can be 1xN or Nx1 vector (where N is number of individuals)
    #pheno_ID: IDs corresponding to pheno.
    #covs: missing values should be encoded as -999.  covs can be 1xN or Nx1 vector, or a NxK matrix (K number of covariates)
    #covs_ID: IDs corresponding to covs
    #kinship_cm: GRM for all mice ever considered (focal or cagemate). !!!contrary to what notation says here, this is all mice = focal + cage mates, not only cm. no missing values allowed. symmetric matrix.
    #kinship_cm_ID: rownames and colnames of kinship_cm
    #cage_cm: cages of all mice ever considered (focal or cagemate). missing values should be encoded as NA. NA cage will lead to this animal to be ignored from the analysis. if you have reason to believe this animal was a cage mate (but you don't know its cage), this is an issue and you need to be aware of it.
    #cage_cm_ID: IDs corresponding to cage_cm
    #independent_covs: if True (default) LIMIX will check whether any covariate is a linear combination of the others and if so fix the issue by updating covs to independent columns
    #DGE: should DGE be included in the model?
    #IGE: should IGE be included in the model?
    #IEE: should IEE be included in the model? note that if IGE = True, IEE will be set to True by the program (silently) - see below
    #cageEffect: should cage effects be included in the model?
    #calc_ste: should the standard errors of the variance components be estimated and output? 
    #standardize_pheno: should the phenotype be standardized to variance 1?
    #subset_IDs: subset of IDs to consider *as focal individuals or cage mate*. this means that any animal not in subset_IDs will be ignored completely from analysis. correct only if all animals from a cage are excluded! if instead want to exclude animals only as focal individuals, set their phenotype to NA.
    def __init__(self, pheno = None, pheno_ID = None, covs = None, covs_ID = None, kinship_cm = None, kinship_cm_ID = None, cage_cm = None, cage_cm_ID = None, independent_covs = True, DGE = True, IGE = False, IEE = False, cageEffect = False, calc_ste = False, standardize_pheno = True, subset_IDs = None):
 
        self.parseNmore(pheno, pheno_ID, covs, covs_ID, kinship_cm, kinship_cm_ID, cage_cm, cage_cm_ID, independent_covs,standardize_pheno, subset_IDs)

        # the purpose of the IEE argument is to specify whether IEE should be off when IGE are off. When IGE are on, IEE must be on and therefore IEE will be automatically set to True when IGE is True.
        if IGE:
            IEE = True
        #VD defines the genetic, environmental and cage covariance matrices
        self.VD(DGE = DGE, IGE = IGE, IEE = IEE, cageEffect = cageEffect)
        #optimize estimates the variance components
        self.optimize()
        # get_VCs gets the variance components, standard errors, etc. for output
        # self.output useful to retrieve output using getOutput without having to specify calc_ste, DGE, IGE, etc.
        self.output = self.get_VCs (calc_ste, DGE = DGE, IGE = IGE, IEE = IEE, cageEffect = cageEffect)
            
    def parseNmore(self, pheno, pheno_ID, covs, covs_ID, kinship_cm, kinship_cm_ID, cage_cm, cage_cm_ID, independent_covs,standardize_pheno, subset_IDs):
        """match various inputs"""

        assert pheno is not None, 'Specify pheno!'
        assert pheno_ID is not None, 'Specify pheno IDs!'
        assert pheno.shape[0] == len(pheno_ID), 'Lengths of pheno and pheno IDs do not match!'
        assert kinship_cm is not None, 'Specify kinship!'
        assert kinship_cm_ID is not None, 'Specify kinship IDs!'
        assert kinship_cm.shape[0] == kinship_cm.shape[1], 'Kinship is not a square matrix!'
        assert kinship_cm.shape[0] == len(kinship_cm_ID), 'Dimension of kinship and length of kinship IDs do not match!'
        assert cage_cm is not None, 'No social analysis possible if cage information is unavailable. For DGE analysis, use regular LIMIX https://github.com/PMBio/limix'

####hack to shorten runtime
#        uCage=sp.unique(cage_cm)
#        remove=uCage[0:510]
#        idx_remove = sp.array(sp.concatenate([sp.where(cage_cm==remove[i])[0] for i in range(len(remove))]))
#        cage_cm[idx_remove]='NA'
####hack to shorten runtime

        #1. define set of animals with cage and kinship information (_cm)
        #1.1 _cm animals need to be in subset_IDs
        if subset_IDs is not None:
            Imatch = sp.nonzero(subset_IDs[:,sp.newaxis]==kinship_cm_ID)
            kinship_cm = kinship_cm[Imatch[1],:][:,Imatch[1]]
            kinship_cm_ID=kinship_cm_ID[Imatch[1]]

        #1.2 NA allowed in cage information so first of all exclude missing cage data and corresponding animals
        has_cage = (cage_cm!='NA')
        if sum(has_cage)==0:
            cage_cm = None
            assert cage_cm is not None, 'No social analysis possible if cage information is unavailable. For DGE analysis, use regular LIMIX https://github.com/PMBio/limix'
        cage_cm=cage_cm[has_cage]
        cage_cm_ID=cage_cm_ID[has_cage]

        #1.3 match cages and kinship as we need both for any social analysis. note that if this is a non social analysis and cage is specified (cage_cm not None), then only those individuals with known cage will be included in the analysis. This is useful to keep same sample between social and non social analyses.
        Imatch = sp.nonzero(cage_cm_ID[:,sp.newaxis]==kinship_cm_ID)
        cage_cm_ID = cage_cm_ID[Imatch[0]]
        cage_cm = cage_cm[Imatch[0]]
        kinship_cm = kinship_cm[Imatch[1],:][:,Imatch[1]]
        kinship_cm_ID=kinship_cm_ID[Imatch[1]]
        #(kinship_cm_ID==cage_cm_ID).all()
        #True
        #cage and kinship now have no missing values and are matched - IDs are in cage_ID and kinship_cm_ID
        # put IDs in sampleID_cm
        sampleID_cm = kinship_cm_ID
        assert len(sampleID_cm)!=0, 'No _cm animals'

        #2. define focal animals now: those with non missing phenotype and non missing covs, kinship and cage, and in subset_IDs
        #2.1 remove NAs from pheno
        if len(pheno.shape)==1:
            pheno = pheno[:,sp.newaxis]
        has_pheno = (pheno!=(-999))[:,0]
        pheno=pheno[has_pheno,:]
        pheno_ID=pheno_ID[has_pheno]
        #2.2 add intercept to covs
        #if no covs are provided, make it a vector of 1s for intercept
        if covs is None:
            covs = sp.ones((pheno.shape[0],1))
            covs_ID = pheno_ID
        #if covs are provided, append a vector of 1s for intercept
        else:
            new_col=sp.ones([covs.shape[0],1])
            if len(covs.shape)==1:
                covs = covs[:,sp.newaxis]
            covs=sp.append(new_col,covs,1)
        #2.3 remove NAs from covs
        has_covs = (covs!=(-999)).all(1)
        covs=covs[has_covs,:]
        covs_ID=covs_ID[has_covs]
        #2.4 match pheno and covs
        Imatch = sp.nonzero(pheno_ID[:,sp.newaxis]==covs_ID)
        pheno = pheno[Imatch[0],:]
        pheno_ID=pheno_ID[Imatch[0]]
        covs = covs[Imatch[1],:]
        covs_ID=covs_ID[Imatch[1]]
        #(pheno_ID==covs_ID).all()
        #True
        #pheno and covs now have no missing values and are matched - IDs are in pheno_ID and covs_ID
        #2.5 check which of those are in sampleID_cm (and thus have kinship and cage)
        has_geno = sp.array([pheno_ID[i] in sampleID_cm for i in range(pheno_ID.shape[0])])
        pheno = pheno[has_geno,:]
        covs = covs[has_geno,:]
        #create sampleID that has focal individuals.
        sampleID=pheno_ID[has_geno]
        assert len(sampleID)!=0, 'No focal animals'
        #remember sampleID_cm and sampleID are in different order (and of different length possibly)

        #3. create cage and kinship for focal animals
        idxs = sp.array([sp.where(sampleID_cm==sampleID[i])[0][0] for i in range(sampleID.shape[0])])
        cage=cage_cm[idxs]
        if len(cage.shape)==1:
            cage = cage[:,sp.newaxis]
        kinship=kinship_cm[idxs,:][:,idxs]
        
        #4. create focal x _cm genetic cross-covariance
        kinship_cross = kinship_cm[idxs,:]
        #so sampleID along rows and sampleID_cm along colummns

        #5. now create environmental matrices
        env = sp.eye(kinship.shape[0])
        env_cm = sp.eye(kinship_cm.shape[0])
        env_cross = env_cm[idxs,:]


        # standardize_pheno should be True for CV if want to output 1-mse
        if standardize_pheno:
            pheno -= pheno.mean(0)
            pheno /= pheno.std(0)
            print('Pheno has been standardized')

        if independent_covs:
            tol = 1e-6
            R = la.qr(covs,mode='r')[0][:covs.shape[1],:]
            I = (abs(R.diagonal())>tol)
            if sp.any(~I):
                warnings.warn('Covariate cols '+str(sp.where(~I)[0])+' have been removed because linearly dependent on the others')
            covs = covs[:,I]

        self.sampleID=sampleID
        self.pheno=pheno
        self.covs=covs
        self.cage=cage
        self.kinship=kinship
        self.env=env
        self.sampleID_cm=sampleID_cm
        self.cage_cm=cage_cm
        self.kinship_cm=kinship_cm
        self.env_cm=env_cm
        self.kinship_cross=kinship_cross
        self.env_cross=env_cross

    def VD(self, DGE, IGE, IEE, cageEffect):
        """ defines covariance for variance decomposition."""

        #defines mean
        mean = lin_mean(self.pheno,self.covs)

        #define cagemate assignment - required for SGE, SEE, and cage effects. Z is N focal x N_cm and has 0s in cells Z_i,i (i.e. an animal is not its own cage mate)
        same_cage = 1. * (self.cage==self.cage_cm)
        diff_inds = 1. * (self.sampleID[:,sp.newaxis]!=self.sampleID_cm)
        Z = same_cage * diff_inds

        #define the overall genetic covariance matrix
        if DGE or IGE:
            #scales kinship (DGE component) to sample variance 1
            sf_K = covar_rescaling_factor(self.kinship)
            self.kinship *= sf_K

            #now create and scale SGE and DGE/SGE covariance components
            if IGE:
                #first SGE component: ZKcmZ' in this code (ZKZ' in paper)
                _ZKcmZ = sp.dot(Z,sp.dot(self.kinship_cm,Z.T))
                sf_ZKcmZ = covar_rescaling_factor(_ZKcmZ)
                self.kinship_cm *= sf_ZKcmZ
                 #second DGE/SGE covariance:
                self.kinship_cross *= sp.sqrt(sf_K * sf_ZKcmZ)
        
        if DGE and not IGE:
            self._genoCov = FixedCov(self.kinship)
        elif IGE and not DGE:
            self._genoCov = FixedCov(_ZKcmZ)
        elif DGE and IGE:
            self._genoCov = DirIndirCov(self.kinship,Z,kinship_cm=self.kinship_cm,kinship_cross=self.kinship_cross)
        else:
            self._genoCov = None


        #define the overall environmental covariance matrix
        #there is always DEE
        #env naturally has sample variance 1 so no need to scale it
        if IEE:
            #_ZZ = ZIcmZ'
            _ZZ  = sp.dot(Z,Z.T)
            sf_ZZ = covar_rescaling_factor(_ZZ)
            self.env_cm *= sf_ZZ
            self.env_cross *= sp.sqrt(1 * sf_ZZ)

            self._envCov = DirIndirCov(self.env,Z,kinship_cm=self.env_cm,kinship_cross=self.env_cross)
        else:
            self._envCov = FixedCov(self.env)

        ##define cage effect covariance matrix
        if cageEffect:
            N = self.pheno.shape[0]
            uCage = sp.unique(self.cage)
            #W, the cage design matrix, is N x n_cages (where N is number of focal animals) 
            W = sp.zeros((N,uCage.shape[0]))
            for cv_i, cv in enumerate(uCage):
                W[:,cv_i] = 1.*(self.cage[:,0]==cv)
            #WW, the cage effect covariance matrix, is N x N and has 1s in cells WW_i,i
            WW = sp.dot(W,W.T)
            #this is equivalent to getting covar_rescaling_factor first and then multiplying, as done for other matrices above
            WW = covar_rescale(WW)
            self._cageCov = FixedCov(WW)
        else:
            self._cageCov = None

        # define overall covariance matrix as sum of genetic, environmental and cage covariance matrices
        if self._genoCov is None:
            if self._cageCov is None:
                self.covar = SumCov(self._envCov)
            else:
                self.covar = SumCov(self._envCov,self._cageCov)
        else:
            if self._cageCov is None:
                self.covar = SumCov(self._genoCov,self._envCov)
            else:
                self.covar = SumCov(self._genoCov,self._envCov,self._cageCov)

        ## define gp
        self._gp = GP(covar=self.covar,mean=mean)
        

    def optimize(self):
        """optimises the covariance matrix = estimate variance components"""
        if 0:
            # trial for inizialization it is complicated though
            cov = sp.array([[0.2,1e-4],[1e-4,1e-4]])
            self._genoCov.setCovariance(cov)
            self._envCov.setCovariance(cov)
            self._cageCov.scale = 0.2
        else:
            self._gp.covar.setRandomParams()

        #optimization - keep calc_ste = False
        self.conv, self.info = self._gp.optimize(calc_ste = False)
        
    def get_VCs(self, calc_ste, DGE, IGE, IEE, cageEffect):
        """function to access estimated variance components, standard errors"""
        if calc_ste:
            try:
                STE_output = self.getGenoSte(DGE, IGE, IEE, cageEffect)
                genetic_STEs = STE_output['R']
                #not sure what corr_params are
                corr_params = STE_output['corr_params']
            
            except:
                genetic_STEs = sp.array([[-999,-999],[-999,-999]])
                corr_params = (-999)
        else:
            genetic_STEs = sp.array([[-999,-999],[-999,-999]])
            corr_params = (-999)

        R = {}
        #whether the run converged
        R['conv'] = self.conv
        #should be small (e.g. < 10^-4)
        R['grad'] = self.info['grad']
        #Contrary to what it says, this is -LML
        R['LML']  = self._gp.LML()
        #number of focal animals
        R['sample_size'] = len(self.sampleID)
        #number of _cm animals. note that this doesnt say much as an animal that is not phenotyped nor a cage mate could still be in there
        R['sample_size_cm'] = len(self.sampleID_cm)
        #standard error for DGE
        R['STE_Ad'] = genetic_STEs[0,0]
        #standard error for SGE
        R['STE_As'] = genetic_STEs[1,1]
        #standard error for covariance between DGE and SGE
        R['STE_Ads'] = genetic_STEs[0,1]
        #covariance between DGE and SGE estimates
        R['corr_params'] = corr_params
        #effect sizes of fixed effects in the model
        R['b'] = self._gp.mean.b

        #variance components and total genetic variance below
        if DGE and (not IGE):
            R['var_Ad'] = self._genoCov.scale
            R['var_As'] = (-999)
            R['corr_Ads'] = (-999)
            R['total_gen_var'] = 1/covar_rescaling_factor(self._genoCov.K())
        elif IGE and (not DGE):
            R['var_Ad'] = (-999)
            R['var_As'] = self._genoCov.scale
            R['corr_Ads'] = (-999)
            R['total_gen_var'] = 1/covar_rescaling_factor(self._genoCov.K())
        elif DGE and IGE:
            R['var_Ad'] = self._genoCov.covff.K()[0,0]
            R['var_As'] = self._genoCov.covff.K()[1,1]
            R['corr_Ads'] = self._genoCov.covff.K()[0,1]/(sp.sqrt(R['var_Ad']) * sp.sqrt(R['var_As']))
            R['total_gen_var'] = 1/covar_rescaling_factor(self._genoCov.K())
        else:
            R['var_Ad'] = (-999)
            R['var_As'] = (-999)
            R['corr_Ads'] = (-999)
            R['total_gen_var'] = (-999)

        if not IEE:
            R['var_Ed'] = self._envCov.scale
            R['var_Es'] = (-999)
            R['corr_Eds'] = (-999)
        else:
            R['var_Ed'] = self._envCov.covff.K()[0,0]
            R['var_Es'] = self._envCov.covff.K()[1,1]
            R['corr_Eds'] = self._envCov.covff.K()[0,1] / (sp.sqrt(R['var_Ed']) * sp.sqrt(R['var_Es']))


        if cageEffect:
            R['var_C'] = self._cageCov.scale
            #environmental covariance matrix (fitted)
            envK = self._envCov.K() + self._cageCov.K()
            #overall (total) covariance matrix (fitted)
            if DGE or IGE:
                totK = self._genoCov.K() + self._envCov.K() + self._cageCov.K()
            else:
                totK = self._envCov.K() + self._cageCov.K()
        else:
            R['var_C'] = (-999)
            envK = self._envCov.K()
            if DGE or IGE:
                totK = self._genoCov.K() + self._envCov.K()
            else:
                totK = self._envCov.K()
        #calculate sample variance of environmental matrix and overall covariance matrices
        R['total_env_var'] = 1/covar_rescaling_factor(envK)
        R['total_var'] = 1/covar_rescaling_factor(totK)
        self.total_variance = R['total_var']

        return R

    def getGenoSte(self, DGE, IGE, IEE, cageEffect):

        self._gp.covar.getFisherInf()
        F = self._gp.covar.getFisherInf()

        # scalar in front of each term
        # ordering for geno and env is
        # direct, covar, indirect as in fisher matrix
        aP = []
        vi = []

        if DGE and (not IGE):
            aP.append(self._genoCov.scale)
            vi.append(1. / covar_rescaling_factor(self._genoCov.K0))
        elif IGE and (not DGE):
            aP.append(self._genoCov.scale)
            vi.append(1. / covar_rescaling_factor(self._genoCov.K0))
        elif DGE and IGE:
            aP.append(self._genoCov.covff.K()[0,0])
            aP.append(self._genoCov.covff.K()[0,1])
            aP.append(self._genoCov.covff.K()[1,1])
            vi.append(1. / covar_rescaling_factor(self._genoCov._K))
            vi.append(1. / covar_rescaling_factor(self._genoCov._KZ + self._genoCov._ZK))
            vi.append(1. / covar_rescaling_factor(self._genoCov._ZKZ))
        else:
            pass

        if not IEE:
            aP.append(self._envCov.scale)
            vi.append(1. / covar_rescaling_factor(self._envCov.K0))
        else:
            aP.append(self._envCov.covff.K()[0,0])
            aP.append(self._envCov.covff.K()[0,1])
            aP.append(self._envCov.covff.K()[1,1])
            vi.append(1. / covar_rescaling_factor(self._envCov._K))
            vi.append(1. / covar_rescaling_factor(self._envCov._KZ + self._envCov._ZK))
            vi.append(1. / covar_rescaling_factor(self._envCov._ZKZ))

        if cageEffect:
            aP.append(self._cageCov.scale)
            vi.append(1. / covar_rescaling_factor(self._cageCov.K0))
        else:
            pass

        # make them vectors
        aP = sp.array(aP)
        vi = sp.array(vi)

        # overall variance
        # this should correspond to the one you get from sampling
        v = (aP*vi).sum()

        # fractions of variance exaplined by each term
        # (can be negative)
        h = (aP*vi) / v

        # jacobean
        J = sp.zeros((aP.shape[0], aP.shape[0]))
        J[:, 0] = h / vi
        J[-1, 1:] = -v / vi[-1]
        for i in range(aP.shape[0]-1):
            J[i, i+1] = v / vi[i]

        # transformation of Fisher
        Fnew = sp.dot(J.T, sp.dot(F, J))

        # invert the new Fisher
        S,U = sp.linalg.eigh(Fnew)
        I = S>1e-9
        U = U[:,I]
        S = S[I]
        FI = sp.dot(U,sp.dot(sp.diag(S**(-1)),U.T))
        # reorder to have same ordering as before
        idxs = list(range(1, aP.shape[0]))
        idxs.append(0)
        FI = FI[idxs, :][:, idxs]
        # R is 2x2 matrix: STE_Ad and STE_As on diag, STE_Ads off
        R = sp.zeros((2, 2))
        STE_output = {}

        if DGE and IGE:
            FI_geno = FI[:3,:][:,:3]
            #STEs = sp.sqrt(FI_geno.diagonal()) ( ordered as Ad Ads As)
            #STEs = sqrt of var of VC corr_params 
            #fills diag and 1 off first
            R[sp.tril_indices(2)] = sp.sqrt(FI_geno.diagonal())
            #now fills other off
            R = R + R.T - sp.diag(R.diagonal())
        
            corr_param_Ad_As = FI_geno[0,2]/(sp.sqrt(FI_geno[0,0])*sp.sqrt(FI_geno[2,2]))
        
        elif DGE and (not IGE):
            R[0,0] = sp.sqrt(FI[0,0])
            R[0,1] = -999
            R[1,0] = -999
            R[1,1] = -999
            corr_param_Ad_As = -999
        
        elif (not DGE) and IGE:
            R[0,0] = -999
            R[0,1] = -999
            R[1,0] = -999
            R[1,1] = sp.sqrt(FI[0,0])
            corr_param_Ad_As = -999
       
        else:
            R[0,0] = -999
            R[0,1] = -999
            R[1,0] = -999
            R[1,1] = -999
            corr_param_Ad_As = -999

        
        STE_output['R']=R
        STE_output['corr_params']= corr_param_Ad_As

        return STE_output


    def getOutput(self):
        """to get output without having to specify DGE, SGE, ...."""
        return self.output
    